import functools
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from pyproj import CRS
from shapely import Geometry
from shapely import box
from shapely import extract_unique_points
from shapely.geometry import Polygon

from ..parallel.parallel import Parallel
from .conversion import to_bbox
from .conversion import to_gdf
from .general import clean_clip
from .general import is_bbox_like


@dataclass
class Gridlooper:
    """Run functions in a loop cellwise based on a grid.

    Args:
        gridsize: Size of the grid cells in units of the crs (meters, degrees).
        mask: Geometry object to create a grid around.
        gridbuffer: Units to buffer each gridcell by. For edge cases.
            Defaults to 0.
        clip: If True (default) geometries are clipped by the grid cells.
            If False, all geometries that intersect will be selected in each iteration.
        verbose: Whether to print progress. Defaults to False.
        keep_geom_type: Whether to keep only the input geometry types after clipping.
            Defaults to True.

    Examples:
    ---------
    Get some points and some polygons.

    >>> import sgis as sg
    >>> points = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")
    >>> points["idx"] = points.index
    >>> buffered = sg.buff(points, 100)
    >>> buffered
        idx                                           geometry
    0      0  POLYGON ((263222.700 6651184.900, 263222.651 6...
    1      1  POLYGON ((272556.100 6653369.500, 272556.051 6...
    2      2  POLYGON ((270182.300 6653032.700, 270182.251 6...
        3      3  POLYGON ((259904.800 6650339.700, 259904.751 6...
    4      4  POLYGON ((272976.200 6652889.100, 272976.151 6...
    ..   ...                                                ...
    995  995  POLYGON ((266901.700 6647844.500, 266901.651 6...
    996  996  POLYGON ((261374.000 6653593.400, 261373.951 6...
    997  997  POLYGON ((263642.900 6645427.000, 263642.851 6...
    998  998  POLYGON ((269326.700 6650628.000, 269326.651 6...
    999  999  POLYGON ((264670.300 6644239.500, 264670.251 6...

    [1000 rows x 2 columns]

    Instantiate a gridlooper.

    >>> looper = sg.Gridlooper(gridsize=200, mask=buffered, concat=True, parallelizer=sg.Parallel(1, backend="multiprocessing"))

    Run the function clean_overlay in a gridloop.

    >>> results = looper.run(
    ...     sg.clean_overlay,
    ...     points,
    ...     buffered,
    ... )
    >>> results
        idx_1 idx_2                        geometry
    0      220   220  POINT (254575.200 6661631.500)
    1      735   735  POINT (256337.400 6649931.700)
    2      575   575  POINT (256369.200 6650413.300)
    3       39    39  POINT (256142.300 6650526.300)
    4      235   235  POINT (256231.300 6650720.200)
    ...    ...   ...                             ...
    1481   711   795  POINT (272845.500 6655048.800)
    1482   711   711  POINT (272845.500 6655048.800)
    1483   757   757  POINT (273507.600 6652806.600)
    1484   457   457  POINT (273524.400 6652979.900)
    1485   284   284  POINT (273650.800 6653000.500)

    [1486 rows x 3 columns]

    """

    gridsize: int
    mask: GeoDataFrame | GeoSeries | Geometry
    gridbuffer: int = 0
    parallelizer: Parallel | None = None
    concat: bool = False
    clip: bool = True
    keep_geom_type: bool = True
    verbose: bool = False

    def __post_init__(self) -> None:
        """Fix types."""
        if not isinstance(self.mask, GeoDataFrame):
            self.mask = to_gdf(self.mask)

    def run(self, func: Callable, *args, **kwargs) -> Any | list[Any]:
        """Run a function for each grid cell in a loop.

        Returns a list of the return values from the function,
        or a (Geo)DataFrame if self.concat is True.

        """

        def _intersects_mask(df: GeoDataFrame) -> pd.Series:
            return df.index.isin(df.sjoin(self.mask).index)

        grid: GeoSeries = (
            make_grid(self.mask, gridsize=self.gridsize).loc[_intersects_mask].geometry
        )

        n = len(grid)

        buffered_grid = grid.buffer(self.gridbuffer, resolution=1, join_style=2)

        if self.parallelizer is not None:
            func_with_clip = functools.partial(
                _clip_and_run_func,
                func=func,
                args=args,
                kwargs=kwargs,
                keep_geom_type=self.keep_geom_type,
                clip=self.clip,
            )
            results = self.parallelizer.map(func_with_clip, buffered_grid)
            if not self.gridbuffer or not self.clip:
                return (
                    results
                    if not self.concat
                    else pd.concat(results, ignore_index=True)
                )
            out = []
            for cell_res, unbuffered in zip(results, grid, strict=True):
                out.append(
                    _clip_back_to_unbuffered_grid(
                        cell_res, unbuffered, self.keep_geom_type
                    )
                )
            return out if not self.concat else pd.concat(out, ignore_index=True)

        results = []
        for i, (unbuffered, buffered) in enumerate(
            zip(grid, buffered_grid, strict=False)
        ):
            cell_kwargs = {
                key: _clip_if_isinstance(
                    value, buffered, self.keep_geom_type, self.clip
                )
                for key, value in kwargs.items()
            }
            cell_args = tuple(
                _clip_if_isinstance(value, buffered, self.keep_geom_type, self.clip)
                for value in args
            )

            cell_res = func(*cell_args, **cell_kwargs)

            # clip back to original
            if self.gridbuffer and self.clip:
                cell_res = _clip_back_to_unbuffered_grid(
                    cell_res, unbuffered, self.keep_geom_type
                )

            results.append(cell_res)

            if self.verbose:
                print(f"Done with {i+1} of {n} grid cells", end="\r")

        return results if not self.concat else pd.concat(results, ignore_index=True)


def gridloop(
    func: Callable,
    gridsize: int,
    mask: GeoDataFrame | GeoSeries | Geometry,
    gridbuffer: int = 0,
    clip: bool = True,
    keep_geom_type: bool = True,
    verbose: bool = False,
    args: tuple | None = None,
    kwargs: dict | None = None,
    parallelizer: Parallel | None = None,
) -> list[Any]:
    """Runs a function in a loop cellwise based on a grid.

    Creates grid from a mask, and runs the function for each cell
    with all GeoDataFrame keyword arguments clipped to the cell
    extent.

    Args:
        func: Function to run cellwise.
        mask: Geometry object to create a grid around.
        gridsize: Size of the grid cells in units of the crs (meters, degrees).
        gridbuffer: Units to buffer each gridcell by. For edge cases.
            Defaults to 0.
        clip: If True (default) geometries are clipped by the grid cells.
            If False, all geometries that intersect will be selected in each iteration.
        verbose: Whether to print progress. Defaults to False.
        keep_geom_type: Whether to keep only the input geometry types after clipping.
            Defaults to True.
        args: Positional arguments to pass to the function. Arguments
            of type GeoDataFrame or GeoSeries will be clipped by the grid cells in
            a loop.
        kwargs: Keyword arguments to pass to the function. Arguments
            of type GeoDataFrame or GeoSeries will be clipped by the grid cells in
            a loop.
        parallelizer: Optional instance of sgis.Parallel, to run the function in parallel.

    Returns:
        List of results with the same length as number of grid cells.

    Raises:
        TypeError: If args or kwargs has a wrong type

    Examples:
    ---------
    Get some points and some polygons.

    >>> import sgis as sg
    >>> points = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")
    >>> points["idx"] = points.index
    >>> buffered = sg.buff(points, 100)
    >>> buffered
         idx                                           geometry
    0      0  POLYGON ((263222.700 6651184.900, 263222.651 6...
    1      1  POLYGON ((272556.100 6653369.500, 272556.051 6...
    2      2  POLYGON ((270182.300 6653032.700, 270182.251 6...
    3      3  POLYGON ((259904.800 6650339.700, 259904.751 6...
    4      4  POLYGON ((272976.200 6652889.100, 272976.151 6...
    ..   ...                                                ...
    995  995  POLYGON ((266901.700 6647844.500, 266901.651 6...
    996  996  POLYGON ((261374.000 6653593.400, 261373.951 6...
    997  997  POLYGON ((263642.900 6645427.000, 263642.851 6...
    998  998  POLYGON ((269326.700 6650628.000, 269326.651 6...
    999  999  POLYGON ((264670.300 6644239.500, 264670.251 6...

    [1000 rows x 2 columns]

    Run the function clean_overlay where the data is clipped to a grid
    of 1000x1000 meters. Args are the first two arguments of clean_overlay,
    kwargs are additional keyword arguments.

    >>> resultslist = sg.gridloop(
    ...     func=sg.clean_overlay,
    ...     mask=buffered,
    ...     gridsize=1000,
    ...     args=(points, buffered),
    ...     kwargs={"how": "intersection"}
    ... )
    >>> type(resultslist)
    list

    >>> results = pd.concat(resultslist, ignore_index=True)
    >>> results
         idx_1 idx_2                        geometry
    0      220   220  POINT (254575.200 6661631.500)
    1      735   735  POINT (256337.400 6649931.700)
    2      575   575  POINT (256369.200 6650413.300)
    3       39    39  POINT (256142.300 6650526.300)
    4      235   235  POINT (256231.300 6650720.200)
    ...    ...   ...                             ...
    1481   711   795  POINT (272845.500 6655048.800)
    1482   711   711  POINT (272845.500 6655048.800)
    1483   757   757  POINT (273507.600 6652806.600)
    1484   457   457  POINT (273524.400 6652979.900)
    1485   284   284  POINT (273650.800 6653000.500)

    [1486 rows x 3 columns]

    """
    if kwargs is None:
        kwargs = {}
    elif not isinstance(kwargs, dict):
        raise TypeError("kwargs should be a dict")

    if args is None:
        args = ()
    elif not isinstance(args, tuple):
        raise TypeError("args should be a tuple")

    if not isinstance(mask, GeoDataFrame):
        mask = to_gdf(mask)

    def _intersects_mask(df: GeoDataFrame) -> pd.Series:
        return df.index.isin(df.sjoin(mask).index)

    grid: GeoSeries = make_grid(mask, gridsize=gridsize).loc[_intersects_mask].geometry

    n = len(grid)

    buffered_grid = grid.buffer(gridbuffer, resolution=1, join_style=2)

    if parallelizer is not None:
        func_with_clip = functools.partial(
            _clip_and_run_func,
            func=func,
            args=args,
            kwargs=kwargs,
            keep_geom_type=keep_geom_type,
            clip=clip,
        )
        results = parallelizer.map(func_with_clip, buffered_grid)
        if not gridbuffer or not clip:
            return results
        out = []
        for cell_res, unbuffered in zip(results, grid, strict=True):
            out.append(
                _clip_back_to_unbuffered_grid(cell_res, unbuffered, keep_geom_type)
            )
        return out

    results = []
    for i, (unbuffered, buffered) in enumerate(zip(grid, buffered_grid, strict=False)):
        cell_kwargs = {
            key: _clip_if_isinstance(value, buffered, keep_geom_type, clip)
            for key, value in kwargs.items()
        }
        cell_args = tuple(
            _clip_if_isinstance(value, buffered, keep_geom_type, clip) for value in args
        )

        cell_res = func(*cell_args, **cell_kwargs)

        # clip back to original
        if gridbuffer and clip:
            cell_res = _clip_back_to_unbuffered_grid(
                cell_res, unbuffered, keep_geom_type
            )

        results.append(cell_res)

        if verbose:
            print(f"Done with {i+1} of {n} grid cells", end="\r")

    return results


def _clip_and_run_func(
    grid_cell: Polygon,
    func: Callable,
    args: tuple,
    kwargs: dict,
    keep_geom_type: bool,
    clip: bool,
) -> Any:
    cell_args = tuple(
        _clip_if_isinstance(value, grid_cell, keep_geom_type, clip) for value in args
    )
    cell_kwargs = {
        key: _clip_if_isinstance(value, grid_cell, keep_geom_type, clip)
        for key, value in kwargs.items()
    }

    return func(*cell_args, **cell_kwargs)


def _clip_if_isinstance(
    value: Any, cell: Geometry, keep_geom_type: bool, clip: bool
) -> Any:
    if not isinstance(value, (gpd.GeoDataFrame | gpd.GeoSeries | Geometry)):
        return value

    if isinstance(value, (gpd.GeoDataFrame | gpd.GeoSeries)):
        if clip:
            return clean_clip(value, cell, keep_geom_type=keep_geom_type)
        return value.loc[value.intersects(cell)]

    return value.intersection(cell).make_valid()


def _clip_back_to_unbuffered_grid(
    results: Any, mask: GeoDataFrame, keep_geom_type: bool
) -> Any:
    if isinstance(results, (gpd.GeoDataFrame | gpd.GeoSeries | Geometry)):
        return _clip_if_isinstance(results, mask, keep_geom_type, clip=True)
    elif isinstance(results, (pd.DataFrame | pd.Series | np.ndarray)):
        return results
    try:
        for key, value in results.items():
            results[key] = _clip_if_isinstance(value, mask, keep_geom_type, clip=True)
    except AttributeError:
        try:
            return [
                _clip_if_isinstance(res, mask, keep_geom_type, clip=True)
                for res in results
            ]
        except TypeError:
            pass
    return results


def make_grid_from_bbox(
    minx: int | float,
    miny: int | float,
    maxx: int | float,
    maxy: int | float,
    *_,
    gridsize: int | float,
    crs: CRS | int | str,
) -> GeoDataFrame:
    """Creates a polygon grid from a bounding box.

    Creates a GeoDataFrame of grid cells of a given size within the given
    maxumum and mimimum x and y values.

    Args:
        minx: Minumum x coordinate.
        miny: Minumum y coordinate.
        maxx: Maximum x coordinate.
        maxy: Maximum y coordinate.
        gridsize: Length of the grid walls.
        crs: Coordinate reference system.

    Returns:
        GeoDataFrame with grid geometries.
    """
    xs0, ys0, xs1, ys1 = [], [], [], []
    for x0 in np.arange(minx, maxx + gridsize, gridsize):
        for y0 in np.arange(miny, maxy + gridsize, gridsize):
            x1 = x0 - gridsize
            y1 = y0 + gridsize
            xs0.append(x0)
            ys0.append(y0)
            xs1.append(x1)
            ys1.append(y1)

    grid_cells = box(xs0, ys0, xs1, ys1)

    return gpd.GeoDataFrame(grid_cells, columns=["geometry"], crs=crs)


def make_grid(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple,
    gridsize: int | float,
    *,
    crs: CRS | None = None,
    clip_to_bounds: bool = False,
) -> GeoDataFrame:
    """Create a polygon grid around geometries.

    Creates a GeoDataFrame of grid cells of a given size around the bounds of
    a given GeoDataFrame. The corners are rounded to the nearest integer.

    Args:
        obj: GeoDataFrame, GeoSeries, shapely geometry or bounding box
            (an iterable with four values (minx, miny, maxx, maxy)).
        gridsize: Length of the grid cell walls.
        crs: Coordinate reference system if 'obj' is not GeoDataFrame or GeoSeries.
        clip_to_bounds: Whether to clip the grid to the total bounds of the geometries.
            Defaults to False.

    Returns:
        GeoDataFrame with grid polygons.

    Raises:
        ValueError: crs can only be None if obj is GeoDataFrame/GeoSeries.

    """
    if isinstance(obj, (GeoDataFrame | GeoSeries)):
        crs = obj.crs or crs
    if hasattr(obj, "__len__") and not len(obj):
        return GeoDataFrame({"geometry": []}, crs=crs)

    minx, miny, maxx, maxy = to_bbox(obj)

    minx = int(minx) if minx > 0 else int(minx - 1)
    miny = int(miny) if miny > 0 else int(miny - 1)

    grid = make_grid_from_bbox(minx, miny, maxx, maxy, gridsize=gridsize, crs=crs)

    if clip_to_bounds:
        grid = grid.clip(to_bbox(obj))

    return grid


def make_ssb_grid(
    gdf: GeoDataFrame | GeoSeries, gridsize: int = 1000, add: int | float = 1
) -> GeoDataFrame:
    """Creates a polygon grid around a GeoDataFrame with an SSB id column.

    Creates a grid that follows the grids produced by Statistics Norway.
    The GeoDataFrame must have 25833 as crs (UTM 33 N).

    Courtesy https://gis.stackexchange.com/questions/269243/creating-polygon-grid-using-geopandas

    Args:
        gdf: A GeoDataFrame.
        gridsize: Size of the grid in meters.
        add: Number of grid cells to add on each side,
            to make sure all data is covered by the grid.

    Returns:
        GeoDataFrame with grid geometries and a column 'SSBID'.

    Raises:
        ValueError: If the GeoDataFrame does not have 25833 as crs.
        TypeError: if gdf has wrong type.
    """
    if not isinstance(gdf, (GeoDataFrame | GeoSeries)):
        raise TypeError("gdf must be GeoDataFrame og GeoSeries.")

    if not gdf.crs.equals(25833):
        raise ValueError(
            "Geodataframe must have crs = 25833. Use df.set_crs(25833) to set "
            "projection or df.to_crs(25833) for transforming."
        )

    minx, miny, maxx, maxy = gdf.total_bounds

    minx = minx - add * gridsize
    miny = miny - add * gridsize
    maxx = maxx + add * gridsize
    maxy = maxy + add * gridsize

    # Adjust for SSB-grid
    if minx > 0:
        minx = int(minx / int(gridsize)) * int(gridsize)
    else:
        minx = int((minx - gridsize) / int(gridsize)) * int(gridsize)

    if minx > 0:
        miny = int(miny / int(gridsize)) * int(gridsize)
    else:
        miny = int((miny - gridsize) / int(gridsize)) * int(gridsize)

    cols = list(np.arange(minx, maxx + gridsize, gridsize))
    rows = list(np.arange(miny, maxy + gridsize, gridsize))

    polygons = []
    for x in cols[:-1]:
        for y in rows[:-1]:
            polygons.append(
                Polygon(
                    [
                        (x, y),
                        (x + gridsize, y),
                        (x + gridsize, y + gridsize),
                        (x, y + gridsize),
                    ]
                )
            )

    grid = gpd.GeoDataFrame({"geometry": polygons}, crs=25833)

    # Make SSB id
    grid["ostc"] = (
        (np.floor((grid.geometry.centroid.x + 2000000) / gridsize) * gridsize).apply(
            int
        )
    ).apply(str)
    grid["nordc"] = (
        (np.floor((grid.geometry.centroid.y) / gridsize) * gridsize).apply(int)
    ).apply(str)
    grid["SSBID"] = grid["ostc"] + grid["nordc"]
    return grid[["SSBID", "geometry"]]


def add_grid_id(
    gdf: GeoDataFrame, gridsize: int, out_column: str = "SSBID"
) -> GeoDataFrame:
    """Adds an SSB grid ID column to a GeoDataFrame of points.

    The GeoDataFrame must have 25833 as crs (UTM 33 N).

    Args:
        gdf: A GeoDataFrame.
        gridsize: Size of the grid in meters.
        out_column: Name of column for the grid id.

    Returns:
        The input GeoDataFrame with a new grid id column.

    Raises:
        ValueError: If the GeoDataFrame does not have 25833 as crs.
    """
    if gdf.crs != 25833:
        raise ValueError(
            "Geodataframe must have crs = 25833. Use df.set_crs(25833) to set "
            "projection or df.to_crs(25833) for transforming."
        )
    midlrdf = gdf.copy()
    midlrdf["ostc"] = (
        (np.floor((midlrdf.geometry.x + 2000000) / gridsize) * gridsize).apply(int)
    ).apply(str)
    midlrdf["nordc"] = (
        (np.floor((midlrdf.geometry.y) / gridsize) * gridsize).apply(int)
    ).apply(str)
    midlrdf[out_column] = midlrdf["ostc"] + midlrdf["nordc"]
    midlrdf = midlrdf.drop(columns=["nordc", "ostc"])
    return midlrdf


def bounds_to_polygon(
    gdf: GeoDataFrame | GeoSeries, copy: bool = True
) -> GeoDataFrame | GeoSeries:
    """Creates a box around the geometry in each row of a GeoDataFrame.

    Args:
        gdf: The GeoDataFrame.
        copy: Defaults to True.

    Returns:
        GeoDataFrame of box polygons with length and index of 'gdf'.

    Examples:
    ---------
    >>> import sgis as sg
    >>> gdf = sg.to_gdf([MultiPoint([(0, 0), (1, 1)]), Point(0, 0)])
    >>> gdf
                                            geometry
    0  MULTIPOINT (0.00000 0.00000, 1.00000 1.00000)
    1                        POINT (0.00000 0.00000)

    >>> sg.bounds_to_polygon(gdf)
                                                geometry
    0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....
    1  POLYGON ((0.00000 0.00000, 0.00000 0.00000, 0....

    """
    if isinstance(gdf, GeoSeries):
        return GeoSeries([box(*arr) for arr in gdf.bounds.values], index=gdf.index)
    if copy:
        gdf = gdf.copy()
    gdf.geometry = [box(*arr) for arr in gdf.bounds.values]
    return gdf


def bounds_to_points(
    gdf: GeoDataFrame | GeoSeries, copy: bool = True
) -> GeoDataFrame | GeoSeries:
    """Creates a 4-noded multipoint around the geometry in each row of a GeoDataFrame.

    Args:
        gdf: The GeoDataFrame.
        copy: Defaults to True.

    Returns:
        GeoDataFrame of multipoints with same length and index as 'gdf'.

    Examples:
    ---------
    >>> import sgis as sg
    >>> gdf = sg.to_gdf([MultiPoint([(0, 0), (1, 1)]), Point(0, 0)])
    >>> gdf
                                            geometry
    0  MULTIPOINT (0.00000 0.00000, 1.00000 1.00000)
    1                        POINT (0.00000 0.00000)

    >>> sg.bounds_to_points(gdf)
                                                geometry
    0  MULTIPOINT (1.00000 0.00000, 1.00000 1.00000, ...
    1                       MULTIPOINT (0.00000 0.00000)
    """
    as_bounds = bounds_to_polygon(gdf, copy=copy)
    if isinstance(gdf, GeoSeries):
        return GeoSeries(extract_unique_points(as_bounds), index=gdf.index)
    gdf.geometry = extract_unique_points(as_bounds.geometry)
    return gdf


def get_total_bounds(
    *geometries: GeoDataFrame | GeoSeries | Geometry, strict: bool = False
) -> tuple[float, float, float, float]:
    """Get a combined total bounds of multiple geometry objects."""
    xs, ys = [], []
    for obj in geometries:
        try:
            minx, miny, maxx, maxy = to_bbox(obj)
            xs += [minx, maxx]
            ys += [miny, maxy]
        except Exception as e:
            try:
                for x in obj:
                    minx, miny, maxx, maxy = to_bbox(x)
                    xs += [minx, maxx]
                    ys += [miny, maxy]
            except Exception as e2:
                if strict:
                    raise e2 from e
                else:
                    continue
    return min(xs), min(ys), max(xs), max(ys)


def points_in_bounds(gdf: GeoDataFrame | GeoSeries, n2: int) -> GeoDataFrame:
    """Get a GeoDataFrame of points within the bounds of the GeoDataFrame."""
    if not isinstance(gdf, (GeoDataFrame | GeoSeries)) and is_bbox_like(gdf):
        minx, miny, maxx, maxy = gdf
    else:
        minx, miny, maxx, maxy = gdf.total_bounds
    xs = np.linspace(minx, maxx, num=n2)
    ys = np.linspace(miny, maxy, num=n2)
    x_coords, y_coords = np.meshgrid(xs, ys, indexing="ij")
    coords = np.concatenate((x_coords.reshape(-1, 1), y_coords.reshape(-1, 1)), axis=1)
    return to_gdf(coords, crs=gdf.crs)



"""Attributes of the legend of the ThematicMap class.

The Legend class is best accessed through the 'legend' attribute of the ThematicMap
class.

"""

import itertools
import warnings
from typing import Any

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from matplotlib.lines import Line2D
from pandas import Series

from ..geopandas_tools.bounds import points_in_bounds

# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


LEGEND_KWARGS = {
    "title",
    "size",
    "position",
    "fontsize",
    "title_fontsize",
    "markersize",
    "framealpha",
    "edgecolor",
    "kwargs",
    "labelspacing",
    "title_color",
    "width",
    "height",
    "labels",
    "pretty_labels",
    "thousand_sep",
    "decimal_mark",
    "label_sep",
    "label_suffix",
    "rounding",
    "facecolor",
    "labelcolor",
}

LOWERCASE_WORDS = {
    "a",
    "an",
    "and",
    "as",
    "at",
    "but",
    "by",
    "for",
    "in",
    "nor",
    "of",
    "on",
    "or",
    "the",
    "up",
}


def prettify_label(label: str) -> str:
    """Replace underscores with spaces and capitalize words that are all lowecase."""
    return " ".join(
        word.title() if word.islower() and word not in LOWERCASE_WORDS else word
        for word in label.replace("_", " ").split()
    )


def prettify_number(x: int | float, rounding: int) -> int:
    rounding = int(float(f"1e+{abs(rounding)}"))
    rounded_down = int(x // rounding * rounding)
    rounded_up = rounded_down + rounding
    diff_up = abs(x - rounded_up)
    diff_down = abs(x - rounded_down)
    if diff_up < diff_down:
        return rounded_up
    else:
        return rounded_down


def prettify_bins(bins: list[int | float], rounding: int) -> list[int]:
    return [
        (
            prettify_number(x, rounding)
            if i != len(bins) - 1
            else int(x)
            # else prettify_number(x, rounding) + abs(rounding)
        )
        for i, x in enumerate(bins)
    ]


class Legend:
    """Holds the general attributes of the legend in the ThematicMap class.

    This class holds attributes of the 'legend' attribute of the ThematicMap class.
    The fontsize, title_fontsize and markersize attributes are adjusted
    according to the size attribute of the ThematicMap.

    If a numeric column is used, additional attributes can be found in the
    ContinousLegend class.

    Examples:
    ---------
    Create ten points with a numeric column from 0 to 9.

    >>> import sgis as sg
    >>> points = sg.to_gdf(
    ...     [
    ...         (0, 1),
    ...         (1, 0),
    ...         (1, 1),
    ...         (0, 0),
    ...         (0.5, 0.5),
    ...         (0.5, 0.25),
    ...         (0.25, 0.25),
    ...         (0.75, 0.75),
    ...         (0.25, 0.75),
    ...         (0.75, 0.25),
    ...     ]
    ... )
    >>> points["number"] = range(10)
    >>> points
                      geometry  number
    0  POINT (0.00000 1.00000)       0
    1  POINT (1.00000 0.00000)       1
    2  POINT (1.00000 1.00000)       2
    3  POINT (0.00000 0.00000)       3
    4  POINT (0.50000 0.50000)       4
    5  POINT (0.50000 0.25000)       5
    6  POINT (0.25000 0.25000)       6
    7  POINT (0.75000 0.75000)       7
    8  POINT (0.25000 0.75000)       8
    9  POINT (0.75000 0.25000)       9

    Creating the ThematicMap instance will also create the legend. Since we
    pass a numeric column, a ContinousLegend is created.

    >>> m = sg.ThematicMap(
    ...     points,
    ...     column="number"
    ...     legend_kwargs=dict(
    ...         title="Meters",
    ...         label_sep="to",
    ...         label_suffix="num",
    ...         rounding=2,
    ...         position = (0.35, 0.28),
    ...         title_fontsize=11,
    ...         fontsize=9,
    ...         markersize=7.5,
    ...     ),
    ... )
    >>> m.plot()
    >>> m.legend
    <sgis.maps.legend.ContinousLegend object at 0x00000222206738D0>
    """

    def __init__(
        self,
        title: str | None = None,
        pretty_labels: bool = True,
        labels: list[str] | None = None,
        position: tuple[float] | None = None,
        markersize: int | None = None,
        fontsize: int | None = None,
        title_fontsize: int | None = None,
        framealpha: float = 1.0,
        edgecolor: str = "#0f0f0f",
        **kwargs,
    ) -> None:
        """Initialiser.

        Args:
            title: Legend title. Defaults to the column name if used in the
                ThematicMap class.
            pretty_labels: If True, words will be capitalized and underscores turned to spaces.
                If continous values, numbers will be rounded.
            labels: Labels of the categories.
            position: The legend's x and y position in the plot, specified as a tuple of
                x and y position between 0 and 1. E.g. position=(0.8, 0.2) for a position
                in the bottom right corner, (0.2, 0.8) for the upper left corner.
            fontsize: Text size of the legend labels. Defaults to the size of
                the ThematicMap class.
            title_fontsize: Text size of the legend title. Defaults to the
                size * 1.2 of the ThematicMap class.
            markersize: Size of the color circles in the legend. Defaults to the size of
                the ThematicMap class.
            framealpha: Transparency of the legend background.
            edgecolor: Color of the legend border. Defaults to #0f0f0f (almost black).
            kwargs: Stores additional keyword arguments taken by the matplotlib legend
                method. Specify this as e.g. m.legend.kwargs["labelcolor"] = "red", where
                'm' is the name of the ThematicMap instance. See here:
                https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html

        """
        self.title = title

        if "size" in kwargs:
            size = kwargs.pop("size")
            self._get_legend_sizes(size, kwargs)
        else:
            self._title_fontsize = title_fontsize
            self._fontsize = fontsize
            self._markersize = markersize

        self.pretty_labels = pretty_labels
        self.framealpha = framealpha
        self.edgecolor = edgecolor
        self.width = kwargs.pop("width", 0.1)
        self.height = kwargs.pop("height", 0.1)
        self.title_color = kwargs.pop("title_color", None)
        self.labelspacing = kwargs.pop("labelspacing", 0.8)

        self.labels = labels
        self._position = position
        self._position_has_been_set = True if position else False

        self.kwargs = {}
        for key, value in kwargs.items():
            if key not in LEGEND_KWARGS:
                self.kwargs[key] = value
            else:
                try:
                    setattr(self, key, value)
                except Exception:
                    setattr(self, f"_{key}", value)

    @property
    def valid_keywords(self) -> set[str]:
        """List all valid keywords for the class initialiser."""
        return LEGEND_KWARGS

    def _get_legend_sizes(self, size: int | float, kwargs: dict) -> None:
        """Adjust fontsize and markersize to size kwarg."""
        if "title_fontsize" in kwargs:
            self._title_fontsize = kwargs["title_fontsize"]
            self._title_fontsize_has_been_set = True
        else:
            self._title_fontsize = size * 1.2

        if "fontsize" in kwargs:
            self._fontsize = kwargs["fontsize"]
            self._fontsize_has_been_set = True
        else:
            self._fontsize = size

        if "markersize" in kwargs:
            self._markersize = kwargs["markersize"]
            self._markersize_has_been_set = True
        else:
            self._markersize = size

    def _prepare_categorical_legend(
        self, categories_colors: dict, nan_label: str
    ) -> None:
        for attr in self.__dict__.keys():
            if attr in self.kwargs:
                self[attr] = self.kwargs.pop(attr)

        # swap column values with label values if labels is dict
        if self.labels and isinstance(self.labels, dict):
            categories_colors = {
                self.labels[cat]: color for cat, color in categories_colors.items()
            }
        # swap column values with label list and hope it's in the correct order
        elif self.labels:
            categories_colors = {
                label: color
                for label, color in zip(
                    self.labels, categories_colors.values(), strict=True
                )
            }

        self._patches, self._categories = [], []
        for category, color in categories_colors.items():
            if self.pretty_labels:
                category = prettify_label(category)
            if category == nan_label:
                self._categories.append(nan_label)
            else:
                self._categories.append(category)
            self._patches.append(
                Line2D(
                    [0],
                    [0],
                    linestyle="none",
                    marker="o",
                    alpha=self.kwargs.get("alpha", 1),
                    markersize=self._markersize,
                    markerfacecolor=color,
                    markeredgewidth=0,
                )
            )

    def _actually_add_legend(self, ax: matplotlib.axes.Axes) -> matplotlib.axes.Axes:
        if self.pretty_labels:
            self.title = prettify_label(self.title)
        legend = ax.legend(
            self._patches,
            self._categories,
            fontsize=self._fontsize,
            title=self.title,
            title_fontsize=self._title_fontsize,
            bbox_to_anchor=self._position + (self.width, self.height),
            fancybox=False,
            framealpha=self.framealpha,
            edgecolor=self.edgecolor,
            labelspacing=self.labelspacing,
            **self.kwargs,
        )

        if self.title_color:
            plt.setp(legend.get_title(), color=self.title_color)

        return ax

    def _get_best_legend_position(
        self, gdf: GeoDataFrame, k: int
    ) -> tuple[float, float]:
        minx, miny, maxx, maxy = gdf.total_bounds
        diffx = maxx - minx
        diffy = maxy - miny

        points = points_in_bounds(gdf, 30)
        gdf = gdf.loc[:, ~gdf.columns.str.contains("index|level_")]
        joined = points.sjoin_nearest(gdf, distance_col="nearest")

        max_distance = max(joined.nearest)

        best_position = joined.loc[joined.nearest == max_distance].drop_duplicates(
            "geometry"
        )

        bestx, besty = (
            best_position.geometry.x.iloc[0],
            best_position.geometry.y.iloc[0],
        )

        bestx_01 = (bestx - minx) / (diffx)
        besty_01 = (besty - miny) / (diffy)

        bestx_01 = 0.1 if bestx_01 < 0.5 else 0.90
        besty_01 = 0.0375 * k if besty_01 < 0.5 else 1

        return bestx_01, besty_01

    def __getitem__(self, item: str) -> Any:
        """Get attribute with square brackets."""
        return getattr(self, item)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set attribute with square brackets."""
        setattr(self, key, value)

    def get(self, key: Any, default: Any = None) -> Any:
        """Get value of an attribute of the Legend."""
        try:
            return self[key]
        except (KeyError, ValueError, IndexError, AttributeError):
            return default

    @property
    def position(self) -> tuple[float, float]:
        """Legend position in x, y."""
        return self._position

    @position.setter
    def position(self, new_value: tuple[float, float]) -> None:
        self._position = new_value
        self._position_has_been_set = True

    @property
    def title_fontsize(self) -> int:
        """Legend title fontsize."""
        return self._title_fontsize

    @title_fontsize.setter
    def title_fontsize(self, new_value: int) -> None:
        self._title_fontsize = new_value
        self._title_fontsize_has_been_set = True

    @property
    def fontsize(self) -> int:
        """Legend fontsize."""
        return self._fontsize

    @fontsize.setter
    def fontsize(self, new_value: int) -> None:
        self._fontsize = new_value
        self._fontsize_has_been_set = True

    @property
    def markersize(self) -> int:
        """Legend markersize."""
        return self._markersize

    @markersize.setter
    def markersize(self, new_value: int) -> None:
        self._markersize = new_value
        self._markersize_has_been_set = True


class ContinousLegend(Legend):
    """Holds the legend attributes specific to numeric columns.

    The attributes consern the labeling of the groups in the legend.
    Labels can be set manually with the 'labels' attribute, or the format
    of the labels can be changed with the remaining attributes.

    Attributes:
        labels: To manually set labels. If set, all other labeling attributes are
            ignored. Should be given as a list of strings with the same length as
            the number of color groups.
        pretty_labels: If False (default), the minimum and maximum values of each
            color group will be used as legend labels. If True, the labels will end
            with the maximum value, but start at 1 + the maximum value of the previous
            group. The labels will be correct but inaccurate.
        label_suffix: The text to put after each number in the legend labels.
            Defaults to None.
        label_sep: Text to put in between the two numbers in each color group in
            the legend. Defaults to '-'.
        thousand_sep: Separator between each thousand for large numbers. Defaults to
            None, meaning no separator.
        decimal_mark: Text to use as decimal point. Defaults to None, meaning '.' (dot)
            unless 'thousand_sep' is '.'. In this case, ',' (comma) will be used as
            decimal mark.

    Examples:
    ---------
    Create ten random points with a numeric column from 0 to 9.

    >>> import sgis as sg
    >>> points = sg.random_points(10)
    >>> points["number"] = range(10)
    >>> points
                    geometry  number
    0  POINT (0.59780 0.50425)       0
    1  POINT (0.07019 0.26167)       1
    2  POINT (0.56475 0.15422)       2
    3  POINT (0.87293 0.60316)       3
    4  POINT (0.47373 0.20040)       4
    5  POINT (0.98661 0.15614)       5
    6  POINT (0.30951 0.77057)       6
    7  POINT (0.47802 0.52824)       7
    8  POINT (0.12215 0.96588)       8
    9  POINT (0.02938 0.93467)       9

    Creating the ThematicMap instance with a numeric column.

    >>> m = sg.ThematicMap(points, column="number")

    Changing the attributes that apply to both numeric and categorical columns.

    >>> m.legend.title = "Meters"
    >>> m.legend.position = (0.35, 0.28)

    Change the attributes that only apply to numeric columns.

    >>> m.label_sep = "to"
    >>> m.label_suffix = "num"
    >>> m.rounding = 2
    >>> m.plot()

    Setting labels manually. For better control, it might be wise to also set the bins
    manually. The following bins will create a plot with the color groups
    0-2, 3-5, 6-7 and 8-9. The legend labels can then be set accordingly.

    >>> m = sg.ThematicMap(points, column="number")
    >>> m.bins = [0, 2, 5, 7, 9]
    >>> m.legend.labels = ["0 to 2 num", "3 to 5 num", "6 to 7 num", "8 to 9 num"]
    >>> m.plot()

    We will get the same groups if we exclude the first and last bin values. The
    minimum and maximum values will be filled anyway.

    >>> m = sg.ThematicMap(points, column="number")
    >>> m.bins = [2, 5, 7]
    >>> m.legend.labels = ["0 to 2 num", "3 to 5 num", "6 to 7 num", "8 to 9 num"]
    >>> m.plot()

    """

    def __init__(
        self,
        labels: list[str] | None = None,
        pretty_labels: bool = True,
        label_suffix: str | None = None,
        label_sep: str = "-",
        rounding: int | None = None,
        thousand_sep: str | None = None,
        decimal_mark: str | None = None,
        **kwargs,
    ) -> None:
        """Initialiser.

        Args:
            labels: To manually set labels. If set, all other labeling attributes are
                ignored. Should be given as a list of strings with the same length as
                the number of color groups.
            pretty_labels: If False (default), the minimum and maximum values of each
                color group will be used as legend labels. If True, the labels will end
                with the maximum value, but start at 1 + the maximum value of the previous
                group. The labels will be correct but inaccurate.
            label_suffix: The text to put after each number in the legend labels.
                Defaults to None.
            label_sep: Text to put in between the two numbers in each color group in
                the legend. Defaults to '-'.
            rounding: Number of decimals in the labels. By default, the rounding
                depends on the column's maximum value and standard deviation.
                OBS: The bins will not be rounded, meaning the labels might be wrong
                if not bins are set manually.
            thousand_sep: Separator between each thousand for large numbers. Defaults to
                None, meaning no separator.
            decimal_mark: Text to use as decimal point. Defaults to None, meaning '.' (dot)
                unless 'thousand_sep' is '.'. In this case, ',' (comma) will be used as
                decimal mark.
            kwargs: Stores additional keyword arguments taken by the matplotlib legend
                method. Specify this as e.g. m.legend.kwargs["labelcolor"] = "red", where
                'm' is the name of the ThematicMap instance. See here:
                https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html

        """
        super().__init__(**kwargs)

        self.pretty_labels = pretty_labels
        self.thousand_sep = thousand_sep
        self.decimal_mark = decimal_mark

        self.label_sep = label_sep
        self.label_suffix = "" if not label_suffix else label_suffix
        self._rounding = rounding
        # self._rounding_has_been_set = True if rounding else False

    def _get_rounding(self, array: Series | np.ndarray) -> int:
        def isinteger(x):
            return np.equal(np.mod(x, 1), 0)

        if np.all(isinteger(array)):
            return 0

        closest_to_zero_idx = np.argmin(np.abs(array))
        closest_to_zero = np.abs(array[closest_to_zero_idx])

        between_1_and_0 = 1 > closest_to_zero > 0
        if between_1_and_0:
            return int(abs(np.log10(abs(closest_to_zero)))) + 1

        std_ = np.std(array)
        max_ = np.max(array)
        if max_ > 30 and std_ > 5:
            return 0
        if max_ > 5 and std_ > 1:
            return 1
        if max_ > 1 and std_ > 0.1:
            return 2
        return int(abs(np.log10(std_))) + 1

    @staticmethod
    def _set_rounding(bins, rounding: int | float) -> list[int | float]:
        if not rounding:
            return [int(round(bin_, 0)) for bin_ in bins]
        elif rounding <= 0:
            return [int(round(bin_, rounding)) for bin_ in bins]
        else:
            return [round(bin_, rounding) for bin_ in bins]

    def _remove_max_legend_value(self) -> None:
        if not self._legend:
            raise ValueError("Cannot modify legend before it is created.")

    def _prepare_continous_legend(
        self,
        bins: list[float],
        colors: list[str],
        nan_label: str,
        bin_values: dict,
    ) -> None:
        # TODO: clean up this messy method

        for attr in self.kwargs:
            if attr in self.__dict__:
                self[attr] = self.kwargs.pop(attr)

        self._patches, self._categories = [], []

        for color in colors:
            self._patches.append(
                Line2D(
                    [0],
                    [0],
                    linestyle="none",
                    marker="o",
                    alpha=self.kwargs.get("alpha", 1),
                    markersize=self._markersize,
                    markerfacecolor=color,
                    markeredgewidth=0,
                )
            )

        if self.labels:
            if len(self.labels) != len(colors):
                raise ValueError(
                    "Label list must be same length as the number of groups. "
                    f"Got k={len(colors)} and labels={len(self.labels)}."
                    f"labels: {', '.join(self.labels)}"
                    f"colors: {', '.join(colors)}"
                    f"bins: {bins}"
                )
            self._categories = self.labels

        elif len(bins) == len(colors):
            for i, _ in enumerate(bins):
                min_ = np.min(bin_values[i])
                max_ = np.max(bin_values[i])
                try:
                    min_rounded = self._set_rounding([min_], self._rounding)[0]
                    max_rounded = self._set_rounding([max_], self._rounding)[0]
                except ValueError as e:
                    if "nan" in str(e).lower():
                        min_rounded, max_rounded == "NaN", "NaN"
                if min_ == max_:
                    self._categories.append(f"{min_rounded} {self.label_suffix}")
                else:
                    self._categories.append(
                        f"{min_rounded} {self.label_suffix} {self.label_sep} "
                        f"{max_rounded} {self.label_suffix}"
                    )

        else:
            for i, (cat1, cat2) in enumerate(itertools.pairwise(bins)):
                if nan_label in str(cat1) or nan_label in str(cat2):
                    self._categories.append(nan_label)
                    continue

                min_ = np.min(bin_values[i])
                max_ = np.max(bin_values[i])

                if self.pretty_labels:
                    if i == 0:
                        cat1 = int(min_) if (self.rounding or 0) <= 0 else min_

                    is_last = i == len(bins) - 2
                    if is_last:
                        cat2 = int(max_) if (self.rounding or 0) <= 0 else max_

                    if (self.rounding or 0) <= 0:
                        cat1 = int(cat1)
                        cat2 = int(cat2 - 1) if not is_last else int(cat2)
                    elif (self.rounding or 0) > 0:
                        cat1 = round(cat1, self._rounding)
                        cat2 = round(
                            cat2 - float(f"1e-{self._rounding}"), self._rounding
                        )
                    else:
                        cat1 = round(cat1, self._rounding)
                        cat2 = round(cat2, self._rounding)

                    cat1 = self._format_number(cat1)
                    cat2 = self._format_number(cat2)

                    if min_ == max_:
                        label = self._get_two_value_label(cat1, cat2)
                        self._categories.append(label)
                        continue

                    label = self._get_two_value_label(cat1, cat2)
                    self._categories.append(label)

                    continue

                min_rounded = self._set_rounding([min_], self._rounding)[0]
                max_rounded = self._set_rounding([max_], self._rounding)[0]
                if min_ == max_:
                    min_rounded = self._format_number(min_rounded)
                    label = self._get_one_value_label(min_rounded)
                    self._categories.append(label)
                else:
                    min_rounded = self._format_number(min_rounded)
                    max_rounded = self._format_number(max_rounded)
                    label = self._get_two_value_label(min_rounded, max_rounded)
                    self._categories.append(label)

    def _get_two_value_label(self, value1: int | float, value2: int | float) -> str:
        return (
            f"{value1} {self.label_suffix} {self.label_sep} "
            f"{value2} {self.label_suffix}"
        )

    def _get_one_value_label(self, value1: int | float) -> str:
        return f"{value1} {self.label_suffix}"

    def _format_number(self, number: int | float) -> int | float:
        if not self.thousand_sep and not self.decimal_mark:
            return number

        if self.thousand_sep:
            number = f"{number:,}".replace(",", "*temp_thousand*")

        if self.decimal_mark:
            number = str(number).replace(".", "*temp_decimal*")

        if self.thousand_sep == "." and not self.decimal_mark:
            number = number.replace(".", ",").replace(
                "*temp_thousand*", self.thousand_sep
            )
        elif not self.thousand_sep:
            number = number.replace("*temp_decimal*", self.decimal_mark)
        elif not self.decimal_mark:
            number = number.replace("*temp_thousand*", self.thousand_sep)
        else:
            number = number.replace("*temp_thousand*", self.thousand_sep).replace(
                "*temp_decimal*", self.decimal_mark
            )
        return number

    @property
    def rounding(self) -> int:
        """Number of decimals in the labels.

        By default, the rounding
        depends on the column's maximum value and standard deviation.
        OBS: The bins will not be rounded, meaning the labels might be wrong
        if not bins are set manually.
        """
        return self._rounding

    @rounding.setter
    def rounding(self, new_value: int) -> None:
        self._rounding = new_value



SENTINEL2_FILENAME_REGEX = r"""
    ^(?P<tile>T\d{2}[A-Z]{3})
    _(?P<date>\d{8})T\d{6}
    # _(?P<date>\d{8}T\d{6})
    _(?P<band>B[018][\dA])
    (?:_(?P<resolution>\d+)m)?
    .*
    \..*$
"""

SENTINEL2_MOSAIC_FILENAME_REGEX = r"""
    ^SENTINEL2X_
    (?P<date>\d{8})
    .*T(?P<tile>\d{2}[A-Z]{3})
    .*(?:_(?P<resolution>{}m))?
    .*(?P<band>B\d{1,2}A|B\d{1,2})
    .*
    .*\..*$
"""


SENTINEL2_CLOUD_FILENAME_REGEX = r"""
    ^(?P<tile>T\d{2}[A-Z]{3})
    _(?P<date>\d{8})T\d{6}
    # _(?P<date>\d{8}T\d{6})
    _(?P<band>SCL)
    (?:_(?P<resolution>\d+)m)?
    .*
    \..*$
"""

SENTINEL2_IMAGE_REGEX = r"""
    ^(?P<mission_id>S2[AB])
    _MSI(?P<level>[A-Z]\d{1}[A-Z])
    _(?P<date>\d{8})T\d{6}
    # _(?P<date>\d{8}T\d{6})
    _(?P<baseline>N\d{4})
    _(?P<orbit>R\d{3})
    _(?P<tile>T\d{2}[A-Z]{3})
    _\d{8}T\d{6}
    .*
    .*$
"""

SENTINEL2_MOSAIC_IMAGE_REGEX = r"""
    ^SENTINEL2X_
    (?P<date>\d{8})
    -\d{6}
    -\d{3}
    _(?P<level>[A-Z]\d{1}[A-Z])
    .*T(?P<tile>\d{2}[A-Z]{3})
    .*.*$
"""

# multiple regex searches because there are different xml files with same info, but different naming
CLOUD_COVERAGE_REGEXES: tuple[str] = (
    r"<Cloud_Coverage_Assessment>([\d.]+)</Cloud_Coverage_Assessment>",
    r"<CLOUDY_PIXEL_OVER_LAND_PERCENTAGE>([\d.]+)</CLOUDY_PIXEL_OVER_LAND_PERCENTAGE>",
)

CRS_REGEX: tuple[str] = (r"<HORIZONTAL_CS_CODE>EPSG:(\d+)</HORIZONTAL_CS_CODE>",)

BOUNDS_REGEX: tuple[dict[str, str]] = (
    {"minx": r"<ULX>(\d+)</ULX>", "maxy": r"<ULY>(\d+)</ULY>"},
)


SENTINEL2_L2A_BANDS = {
    "B01": 60,
    "B02": 10,
    "B03": 10,
    "B04": 10,
    "B05": 20,
    "B06": 20,
    "B07": 20,
    "B08": 10,
    "B8A": 20,
    "B09": 60,
    "B11": 20,
    "B12": 20,
}
SENTINEL2_L1C_BANDS = SENTINEL2_L2A_BANDS | {"B10": 60}
SENTINEL2_CLOUD_BANDS = {
    "SCL": 20,  # SCL: scene classification
}

SENTINEL2_SCL_CLASSES = {
    0: "No Data (Missing data)",  # 000000
    1: "Saturated or defective pixel",  # ff0000
    2: "Topographic casted shadows",  # 2f2f2f
    3: "Cloud shadows",  # 643200
    4: "Vegetation",  # 00a000
    5: "Not-vegetated",  # ffe65a
    6: "Water",  # 0000ff
    7: "Unclassified",  # 808080
    8: "Cloud medium probability",  # c0c0c0
    9: "Cloud high probability",  # ffffff
    10: "Thin cirrus",  # 64c8ff
    11: "Snow or ice",  # ff96ff
}

SENTINEL2_BANDS = SENTINEL2_L1C_BANDS | SENTINEL2_CLOUD_BANDS
SENTINEL2_RBG_BANDS = ["B02", "B03", "B04"]
SENTINEL2_NDVI_BANDS = ["B04", "B08"]



import warnings

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely import Geometry

from .conversion import to_gdf

gdf_type_error_message = "'gdf' should be of type GeoDataFrame or GeoSeries."


def sfilter(
    gdf: GeoDataFrame | GeoSeries,
    other: GeoDataFrame | GeoSeries | Geometry,
    predicate: str = "intersects",
) -> GeoDataFrame:
    """Filter a GeoDataFrame or GeoSeries by spatial predicate.

    Does an sjoin and returns the rows of 'gdf' that were returned
    without getting duplicates or columns from 'other'.
    Works with unique and non-unique index.

    Like 'select by location' in ArcGIS/QGIS, except that the
    selection is permanent.

    Args:
        gdf: The GeoDataFrame.
        other: The geometry object to filter 'gdf' by.
        predicate: Spatial predicate to use. Defaults to 'intersects'.

    Returns:
        A copy of 'gdf' with only the rows matching the
        spatial predicate with 'other'.

    Examples:
    ---------
    >>> import sgis as sg
    >>> df1 = sg.to_gdf([(0, 0), (0, 1)])
    >>> df1
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (0.00000 1.00000)
    >>> df2 = sg.to_gdf([(0, 0), (1, 2)])
    >>> df2
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (1.00000 2.00000)

    Keep rows in df1 intersecting any geometry in df2.

    >>> sg.sfilter(df1, df2)
                      geometry
    0  POINT (0.00000 0.00000)

    Equivelent to sjoin-ing and selecting based on integer index
    (in case of non-unique index).

    >>> df1["idx"] = range(len(df1))
    >>> joined = df1.sjoin(df2)
    >>> df1.loc[df1["idx"].isin(joined["idx"])].drop(columns="idx")
                          geometry
    0  POINT (0.00000 0.00000)

    Also equivelent to using the intersects method, which
    is often a lot slower since df2 must be dissolved:

    >>> df1.loc[df1.intersects(df2.unary_union)]
                      geometry
    0  POINT (0.00000 0.00000)

    """
    if not isinstance(gdf, (GeoDataFrame | GeoSeries)):
        raise TypeError(gdf_type_error_message)

    other = _sfilter_checks(other, crs=gdf.crs)

    indices = _get_sfilter_indices(gdf, other, predicate)

    return gdf.iloc[indices]


def sfilter_split(
    gdf: GeoDataFrame | GeoSeries,
    other: GeoDataFrame | GeoSeries | Geometry,
    predicate: str = "intersects",
) -> tuple[GeoDataFrame, GeoDataFrame]:
    """Split a GeoDataFrame or GeoSeries by spatial predicate.

    Like sfilter, but returns both the rows that do and do not match
    the spatial predicate as separate GeoDataFrames.

    Args:
        gdf: The GeoDataFrame.
        other: The geometry object to filter 'gdf' by.
        predicate: Spatial predicate to use. Defaults to 'intersects'.

    Returns:
        A tuple of GeoDataFrames, one with the rows that match the spatial predicate
        and one with the rows that do not.

    Examples:
    ---------
    >>> import sgis as sg
    >>> df1 = sg.to_gdf([(0, 0), (0, 1)])
    >>> df1
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (0.00000 1.00000)
    >>> df2 = sg.to_gdf([(0, 0), (1, 2)])
    >>> df2
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (1.00000 2.00000)

    Split df1 into the rows that do and do not intersect df2.

    >>> intersecting, not_intersecting = sg.sfilter_split(df1, df2)
    >>> intersecting
                      geometry
    0  POINT (0.00000 0.00000)
    >>> not_intersecting
                      geometry
    1  POINT (0.00000 1.00000)

    Equivelent to sjoin-ing and selecting based on index (which requires the
    index to be unique).

    >>> df1 = df1.reset_index(drop=True)
    >>> joined = df1.sjoin(df2)
    >>> intersecting = df1.loc[df1.index.isin(joined.index)]
    >>> not_intersecting = df1.loc[~df1.index.isin(joined.index)]

    Also equivelent to using the intersects method, which
    is often slower since df2 must be dissolved:

    >>> filt = df1.intersects(df2.unary_union)
    >>> intersecting = df1.loc[filt]
    >>> not_intersecting = df1.loc[~filt]

    """
    if not isinstance(gdf, (GeoDataFrame | GeoSeries)):
        raise TypeError(gdf_type_error_message)

    other = _sfilter_checks(other, crs=gdf.crs)

    indices = _get_sfilter_indices(gdf, other, predicate)

    return (
        gdf.iloc[indices],
        gdf.iloc[pd.Index(range(len(gdf))).difference(pd.Index(indices))],
    )


def sfilter_inverse(
    gdf: GeoDataFrame | GeoSeries,
    other: GeoDataFrame | GeoSeries | Geometry,
    predicate: str = "intersects",
) -> GeoDataFrame | GeoSeries:
    """Filter a GeoDataFrame or GeoSeries by inverse spatial predicate.

    Returns the rows that do not match the spatial predicate.

    Args:
        gdf: The GeoDataFrame or GeoSeries.
        other: The geometry object to filter 'gdf' by.
        predicate: Spatial predicate to use. Defaults to 'intersects'.

    Returns:
        A copy of 'gdf' with only the rows that do not match the
        spatial predicate with 'other'.

    Examples:
    ---------
    >>> import sgis as sg
    >>> df1 = sg.to_gdf([(0, 0), (0, 1)])
    >>> df1
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (0.00000 1.00000)
    >>> df2 = sg.to_gdf([(0, 0), (1, 2)])
    >>> df2
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (1.00000 2.00000)

    Keep the rows in df1 that do not intersect a geometry in df2.

    >>> not_intersecting = sg.sfilter_inverse(df1, df2)
    >>> not_intersecting
                      geometry
    1  POINT (0.00000 1.00000)

    Equivelent to sjoin-ing and selecting based on index (which requires the
    index to be unique).

    >>> df1 = df1.reset_index(drop=True)
    >>> joined = df1.sjoin(df2)
    >>> not_intersecting = df1.loc[~df1.index.isin(joined.index)]

    Also equivelent to using the intersects method, which
    is often slower since df2 must be dissolved:

    >>> not_intersecting = df1.loc[~df1.intersects(df2.unary_union)]

    """
    if not isinstance(gdf, (GeoDataFrame | GeoSeries)):
        raise TypeError(gdf_type_error_message)

    other = _sfilter_checks(other, crs=gdf.crs)

    indices = _get_sfilter_indices(gdf, other, predicate)

    return gdf.iloc[pd.Index(range(len(gdf))).difference(pd.Index(indices))]


def _sfilter_checks(other, crs):
    """Allow 'other' to be any geometry object."""
    if isinstance(other, GeoSeries):
        return other

    if not isinstance(other, GeoDataFrame):
        try:
            other = to_gdf(other)
        except TypeError as e:
            raise TypeError(
                f"Unexpected type of 'other' {other.__class__.__name__}"
            ) from e

        if crs is None:
            return other

        try:
            other = other.set_crs(crs)
        except ValueError as e:
            raise ValueError("crs mismatch", crs, other.crs) from e

    return other


def _get_sfilter_indices(
    left: GeoDataFrame | GeoSeries,
    right: GeoDataFrame | GeoSeries | Geometry,
    predicate: str,
) -> np.ndarray:
    """Compute geometric comparisons and get matching indices.

    Taken from:
    geopandas.tools.sjoin._geom_predicate_query

    Parameters
    ----------
    left : GeoDataFrame
    right : GeoDataFrame
    predicate : string
        Binary predicate to query.

    Returns:
    -------
    DataFrame
        DataFrame with matching indices in
        columns named `_key_left` and `_key_right`.
    """
    original_predicate = predicate

    with warnings.catch_warnings():
        # We don't need to show our own warning here
        # TODO remove this once the deprecation has been enforced
        warnings.filterwarnings(
            "ignore", "Generated spatial index is empty", FutureWarning
        )

        if predicate == "within":
            # within is implemented as the inverse of contains
            # contains is a faster predicate
            # see discussion at https://github.com/geopandas/geopandas/pull/1421
            predicate = "contains"
            sindex = left.sindex
            input_geoms = right.geometry if isinstance(right, GeoDataFrame) else right
        else:
            # all other predicates are symmetric
            # keep them the same
            sindex = right.sindex
            input_geoms = left.geometry if isinstance(left, GeoDataFrame) else left

    l_idx, r_idx = sindex.query(input_geoms, predicate=predicate, sort=False)

    if original_predicate == "within":
        return np.sort(np.unique(r_idx))

    return np.sort(np.unique(l_idx))



import numbers
import warnings
from collections.abc import Hashable
from collections.abc import Iterable
from typing import Any

import joblib
import numpy as np
import pandas as pd
import pyproj
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from geopandas.array import GeometryArray
from geopandas.array import GeometryDtype
from numpy.typing import NDArray
from shapely import Geometry
from shapely import get_coordinates
from shapely import get_exterior_ring
from shapely import get_interior_ring
from shapely import get_num_interior_rings
from shapely import get_parts
from shapely import linestrings
from shapely import make_valid
from shapely import points as shapely_points
from shapely import unary_union
from shapely.geometry import LineString
from shapely.geometry import Point

try:
    import dask_geopandas
except ImportError:
    pass

from .geometry_types import get_geom_type
from .geometry_types import make_all_singlepart
from .geometry_types import to_single_geom_type


def split_geom_types(gdf: GeoDataFrame | GeoSeries) -> tuple[GeoDataFrame | GeoSeries]:
    return tuple(
        gdf.loc[gdf.geom_type == geom_type] for geom_type in gdf.geom_type.unique()
    )


def get_common_crs(
    iterable: Iterable[Hashable], strict: bool = False
) -> pyproj.CRS | None:
    """Returns the common not-None crs or raises a ValueError if more than one.

    Args:
        iterable: Iterable of objects with the attribute "crs" or a list
            of CRS-like (pyproj.CRS-accepted) objects.
        strict: If False (default), falsy CRS-es will be ignored and None
            will be returned if all CRS-es are falsy. If strict is True,

    Returns:
        pyproj.CRS object or None (if all crs are None).

    Raises:
        ValueError if there are more than one crs. If strict is True,
        None is included.
    """
    crs = set()
    for obj in iterable:
        try:
            crs.add(obj.crs)
        except AttributeError:
            pass

    if not crs:
        try:
            crs = list(set(iterable))
        except TypeError:
            return None

    truthy_crs = list({x for x in crs if x})

    if strict and len(truthy_crs) != len(crs):
        raise ValueError("Mix of falsy and truthy CRS-es found.")

    if len(truthy_crs) > 1:
        # sometimes the bbox is slightly different, resulting in different
        # hash values for same crs. Therefore, trying to
        actually_different = set()
        for x in truthy_crs:
            if x.to_string() in {j.to_string() for j in actually_different}:
                continue
            actually_different.add(x)

        if len(actually_different) == 1:
            return next(iter(actually_different))
        raise ValueError("'crs' mismatch.", truthy_crs)

    return pyproj.CRS(truthy_crs[0])


def is_bbox_like(obj: Any) -> bool:
    if (
        hasattr(obj, "__iter__")
        and len(obj) == 4
        and all(isinstance(x, numbers.Number) for x in obj)
    ):
        return True
    return False


def is_wkt(text: str) -> bool:
    gemetry_types = ["point", "polygon", "line", "geometrycollection"]
    return any(x in text.lower() for x in gemetry_types)


def _push_geom_col(gdf: GeoDataFrame) -> GeoDataFrame:
    """Makes the geometry column the rightmost column in the GeoDataFrame.

    Args:
        gdf: GeoDataFrame.

    Returns:
        The GeoDataFrame with the geometry column pushed all the way to the right.
    """
    geom_col = gdf._geometry_column_name
    return gdf.reindex(columns=[c for c in gdf.columns if c != geom_col] + [geom_col])


def drop_inactive_geometry_columns(gdf: GeoDataFrame) -> GeoDataFrame:
    """Removes geometry columns in a GeoDataFrame if they are not active."""
    for col in gdf.columns:
        if (
            isinstance(gdf[col].dtype, GeometryDtype)
            and col != gdf._geometry_column_name
        ):
            gdf = gdf.drop(col, axis=1)
    return gdf


def _rename_geometry_if(gdf: GeoDataFrame) -> GeoDataFrame:
    geom_col = gdf._geometry_column_name
    if geom_col == "geometry" and geom_col in gdf.columns:
        return gdf
    elif geom_col in gdf.columns:
        return gdf.rename_geometry("geometry")

    geom_cols = list(
        {col for col in gdf.columns if isinstance(gdf[col].dtype, GeometryDtype)}
    )
    if len(geom_cols) == 1:
        gdf._geometry_column_name = geom_cols[0]
        return gdf.rename_geometry("geometry")

    raise ValueError(
        "There are multiple geometry columns and none are the active geometry"
    )


def clean_geoms(
    gdf: GeoDataFrame | GeoSeries,
    ignore_index: bool = False,
) -> GeoDataFrame | GeoSeries:
    """Fixes geometries, then removes empty, NaN and None geometries.

    Args:
        gdf: GeoDataFrame or GeoSeries to be cleaned.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False

    Returns:
        GeoDataFrame or GeoSeries with fixed geometries and only the rows with valid,
        non-empty and not-NaN/-None geometries.

    Examples:
    ---------
    >>> import sgis as sg
    >>> import pandas as pd
    >>> from shapely import wkt
    >>> gdf = sg.to_gdf([
    ...         "POINT (0 0)",
    ...         "LINESTRING (1 1, 2 2)",
    ...         "POLYGON ((3 3, 4 4, 3 4, 3 3))"
    ...         ])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    Add None and empty geometries.

    >>> missing = pd.DataFrame({"geometry": [None]})
    >>> empty = sg.to_gdf(wkt.loads("POINT (0 0)").buffer(0))
    >>> gdf = pd.concat([gdf, missing, empty])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....
    0                                               None
    0                                      POLYGON EMPTY

    Clean.

    >>> sg.clean_geoms(gdf)
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....
    """
    warnings.filterwarnings("ignore", "GeoSeries.notna", UserWarning)

    if isinstance(gdf, GeoDataFrame):
        # only repair if necessary
        if not gdf.geometry.is_valid.all():
            gdf.geometry = gdf.make_valid()

        notna = gdf.geometry.notna()
        if not notna.all():
            gdf = gdf.loc[notna]

        is_empty = gdf.geometry.is_empty
        if is_empty.any():
            gdf = gdf.loc[~is_empty]

    elif isinstance(gdf, GeoSeries):
        if not gdf.is_valid.all():
            gdf = gdf.make_valid()

        notna = gdf.notna()
        if not notna.all():
            gdf = gdf.loc[notna]

        is_empty = gdf.is_empty
        if is_empty.any():
            gdf = gdf.loc[~is_empty]

    else:
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    if ignore_index:
        gdf = gdf.reset_index(drop=True)

    return gdf


def get_grouped_centroids(
    gdf: GeoDataFrame, groupby: str | list[str], as_string: bool = True
) -> pd.Series:
    """Get the centerpoint of the geometries within a group.

    Args:
        gdf: GeoDataFrame.
        groupby: column to group by.
        as_string: If True (default), coordinates are returned in
            the format "{x}_{y}". If False, coordinates are returned
            as Points.

    Returns:
        A pandas.Series of grouped centroids with the index of 'gdf'.
    """
    centerpoints = gdf.assign(geometry=lambda x: x.centroid)

    grouped_centerpoints = centerpoints.dissolve(by=groupby).assign(
        geometry=lambda x: x.centroid
    )
    xs = grouped_centerpoints.geometry.x
    ys = grouped_centerpoints.geometry.y

    if as_string:
        grouped_centerpoints["wkt"] = [
            f"{int(x)}_{int(y)}" for x, y in zip(xs, ys, strict=False)
        ]
    else:
        grouped_centerpoints["wkt"] = [
            Point(x, y) for x, y in zip(xs, ys, strict=False)
        ]

    return gdf[groupby].map(grouped_centerpoints["wkt"])


def sort_large_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by area in decending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from large to small in area.

    Examples:
    ---------
    Create GeoDataFrame with NaN values.

    >>> import sgis as sg
    >>> df = sg.to_gdf(
    ...     [
    ...         (0, 1),
    ...         (1, 0),
    ...         (1, 1),
    ...         (0, 0),
    ...         (0.5, 0.5),
    ...     ]
    ... )
    >>> df.geometry = df.buffer([4, 1, 2, 3, 5])
    >>> df["col"] = [None, 1, 2, None, 1]
    >>> df["col2"] = [None, 1, 2, 3, None]
    >>> df["area"] = df.area
    >>> df
                                                geometry  col  col2       area
    0  POLYGON ((4.56136 0.53436, 4.54210 0.14229, 4....  NaN   NaN  50.184776
    1  POLYGON ((1.40111 0.71798, 1.39630 0.61996, 1....  1.0   1.0   3.136548
    2  POLYGON ((2.33302 0.49287, 2.32339 0.29683, 2....  2.0   2.0  12.546194
    3  POLYGON ((3.68381 0.46299, 3.66936 0.16894, 3....  NaN   3.0  28.228936
    4  POLYGON ((5.63590 0.16005, 5.61182 -0.33004, 5...  1.0   NaN  78.413712

    >>> sg.sort_large_first(df)
                                                geometry  col  col2       area
    4  POLYGON ((5.63590 0.16005, 5.61182 -0.33004, 5...  1.0   NaN  78.413712
    0  POLYGON ((4.56136 0.53436, 4.54210 0.14229, 4....  NaN   NaN  50.184776
    3  POLYGON ((3.68381 0.46299, 3.66936 0.16894, 3....  NaN   3.0  28.228936
    2  POLYGON ((2.33302 0.49287, 2.32339 0.29683, 2....  2.0   2.0  12.546194
    1  POLYGON ((1.40111 0.71798, 1.39630 0.61996, 1....  1.0   1.0   3.136548

    >>> sg.sort_nans_last(sg.sort_large_first(df))
                                                geometry  col  col2       area
    2  POLYGON ((2.33302 0.49287, 2.32339 0.29683, 2....  2.0   2.0  12.546194
    1  POLYGON ((1.40111 0.71798, 1.39630 0.61996, 1....  1.0   1.0   3.136548
    4  POLYGON ((5.63590 0.16005, 5.61182 -0.33004, 5...  1.0   NaN  78.413712
    3  POLYGON ((3.68381 0.46299, 3.66936 0.16894, 3....  NaN   3.0  28.228936
    0  POLYGON ((4.56136 0.53436, 4.54210 0.14229, 4....  NaN   NaN  50.184776
    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    area_mapper = dict(enumerate(gdf.area.values))
    sorted_areas = dict(reversed(sorted(area_mapper.items(), key=lambda item: item[1])))
    return gdf.iloc[list(sorted_areas)]


def sort_long_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by length in decending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from long to short in length.
    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    length_mapper = dict(enumerate(gdf.length.values))
    sorted_lengths = dict(
        reversed(sorted(length_mapper.items(), key=lambda item: item[1]))
    )
    return gdf.iloc[list(sorted_lengths)]


def sort_short_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by length in ascending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from short to long in length.
    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    length_mapper = dict(enumerate(gdf.length.values))
    sorted_lengths = dict(sorted(length_mapper.items(), key=lambda item: item[1]))
    return gdf.iloc[list(sorted_lengths)]


def sort_small_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by area in ascending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from small to large in area.

    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    area_mapper = dict(enumerate(gdf.area.values))
    sorted_areas = dict(sorted(area_mapper.items(), key=lambda item: item[1]))
    return gdf.iloc[list(sorted_areas)]


def make_lines_between_points(
    arr1: NDArray[Point] | GeometryArray | GeoSeries,
    arr2: NDArray[Point] | GeometryArray | GeoSeries,
) -> NDArray[LineString]:
    """Creates an array of linestrings from two arrays of points.

    The operation is done rowwise.

    Args:
        arr1: GeometryArray og GeoSeries of points.
        arr2: GeometryArray og GeoSeries of points of same length as arr1.

    Returns:
        A numpy array of linestrings.

    Raises:
        ValueError: If the arrays have unequal shape.

    """
    if arr1.shape != arr2.shape:
        raise ValueError("Arrays must have equal shape.")

    coords: pd.DataFrame = pd.concat(
        [
            pd.DataFrame(get_coordinates(arr1), columns=["x", "y"]),
            pd.DataFrame(get_coordinates(arr2), columns=["x", "y"]),
        ]
    ).sort_index()

    return linestrings(coords.values, indices=coords.index)


def random_points(n: int, loc: float | int = 0.5) -> GeoDataFrame:
    """Creates a GeoDataFrame with n random points.

    Args:
        n: Number of points/rows to create.
        loc: Mean ('centre') of the distribution.

    Returns:
        A GeoDataFrame of points with n rows.

    Examples:
    ---------
    >>> import sgis as sg
    >>> points = sg.random_points(10_000)
    >>> points
                         geometry
    0     POINT (0.62044 0.22805)
    1     POINT (0.31885 0.38109)
    2     POINT (0.39632 0.61130)
    3     POINT (0.99401 0.35732)
    4     POINT (0.76403 0.73539)
    ...                       ...
    9995  POINT (0.90433 0.75080)
    9996  POINT (0.10959 0.59785)
    9997  POINT (0.00330 0.79168)
    9998  POINT (0.90926 0.96215)
    9999  POINT (0.01386 0.22935)
    [10000 rows x 1 columns]

    Values with a mean of 100.

    >>> points = sg.random_points(10_000, loc=100)
    >>> points
                         geometry
    0      POINT (50.442 199.729)
    1       POINT (26.450 83.367)
    2     POINT (111.054 147.610)
    3      POINT (93.141 141.456)
    4       POINT (94.101 24.837)
    ...                       ...
    9995   POINT (174.344 91.772)
    9996    POINT (95.375 11.391)
    9997    POINT (45.694 60.843)
    9998   POINT (73.261 101.881)
    9999  POINT (134.503 168.155)
    [10000 rows x 1 columns]
    """
    if isinstance(n, (str, float)):
        n = int(n)

    x = np.random.rand(n) * float(loc) * 2
    y = np.random.rand(n) * float(loc) * 2

    return GeoDataFrame(
        (Point(x, y) for x, y in zip(x, y, strict=True)), columns=["geometry"]
    )


def random_points_in_polygons(gdf: GeoDataFrame, n: int, seed=None) -> GeoDataFrame:
    """Creates a GeoDataFrame with n random points within the geometries of 'gdf'.

    Args:
        gdf: A GeoDataFrame.
        n: Number of points/rows to create.
        seed: Optional random seet.

    Returns:
        A GeoDataFrame of points with n rows.
    """
    all_points = []

    rng = np.random.default_rng(seed)

    for i, geom in enumerate(gdf.geometry):
        minx, miny, maxx, maxy = geom.bounds

        xs = rng.uniform(minx, maxx, size=n * 500)
        ys = rng.uniform(miny, maxy, size=n * 500)

        points = GeoSeries(shapely_points(xs, y=ys), index=[i] * len(xs))
        all_points.append(points)

    return (
        pd.concat(all_points)
        .loc[lambda x: x.intersects(gdf.geometry)]
        .groupby(level=0)
        .head(n)
    )


def to_lines(*gdfs: GeoDataFrame, copy: bool = True) -> GeoDataFrame:
    """Makes lines out of one or more GeoDataFrames and splits them at intersections.

    The GeoDataFrames' geometries are converted to LineStrings, then unioned together
    and made to singlepart. The lines are split at the intersections. Mimics
    'feature to line' in ArcGIS.

    Args:
        *gdfs: one or more GeoDataFrames.
        copy: whether to take a copy of the incoming GeoDataFrames. Defaults to True.

    Returns:
        A GeoDataFrame with singlepart line geometries and columns of all input
            GeoDataFrames.

    Note:
        The index is preserved if only one GeoDataFrame is given, but otherwise
        ignored. This is because the union overlay used if multiple GeoDataFrames
        always ignores the index.

    Examples:
    ---------
    Convert single polygon to linestring.

    >>> import sgis as sg
    >>> from shapely.geometry import Polygon
    >>> poly1 = sg.to_gdf(Polygon([(0, 0), (0, 1), (1, 1), (1, 0)]))
    >>> poly1["poly1"] = 1
    >>> line = sg.to_lines(poly1)
    >>> line
                                                geometry  poly1
    0  LINESTRING (0.00000 0.00000, 0.00000 1.00000, ...      1

    Convert two overlapping polygons to linestrings.

    >>> poly2 = sg.to_gdf(Polygon([(0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5)]))
    >>> poly2["poly2"] = 1
    >>> lines = sg.to_lines(poly1, poly2)
    >>> lines
    poly1  poly2                                           geometry
    0    1.0    NaN  LINESTRING (0.00000 0.00000, 0.00000 1.00000, ...
    1    1.0    NaN  LINESTRING (0.50000 1.00000, 1.00000 1.00000, ...
    2    1.0    NaN  LINESTRING (1.00000 0.50000, 1.00000 0.00000, ...
    3    NaN    1.0      LINESTRING (0.50000 0.50000, 0.50000 1.00000)
    4    NaN    1.0  LINESTRING (0.50000 1.00000, 0.50000 1.50000, ...
    5    NaN    1.0      LINESTRING (1.00000 0.50000, 0.50000 0.50000)

    Plot before and after.

    >>> sg.qtm(poly1, poly2)
    >>> lines["l"] = lines.length
    >>> sg.qtm(lines, "l")
    """
    if not all(isinstance(gdf, (GeoSeries, GeoDataFrame)) for gdf in gdfs):
        raise TypeError("gdf must be GeoDataFrame or GeoSeries")

    if any(gdf.geom_type.isin(["Point", "MultiPoint"]).any() for gdf in gdfs):
        raise ValueError("Cannot convert points to lines.")

    def _shapely_geometry_to_lines(geom):
        """Get all lines from the exterior and interiors of a Polygon."""
        # if lines (points are not allowed in this function)
        if geom.area == 0:
            return geom

        singlepart = get_parts(geom)
        lines = []
        for part in singlepart:
            exterior_ring = get_exterior_ring(part)
            lines.append(exterior_ring)

            n_interior_rings = get_num_interior_rings(part)
            if not (n_interior_rings):
                continue

            interior_rings = [
                LineString(get_interior_ring(part, n)) for n in range(n_interior_rings)
            ]

            lines += interior_rings

        return unary_union(lines)

    lines = []
    for gdf in gdfs:
        if copy:
            gdf = gdf.copy()

        mapped = gdf.geometry.map(_shapely_geometry_to_lines)
        try:
            gdf.geometry = mapped
        except AttributeError:
            # geoseries
            gdf.loc[:] = mapped

        gdf = to_single_geom_type(gdf, "line")

        lines.append(gdf)

    if len(lines) == 1:
        return lines[0]

    if len(lines[0]) and len(lines[1]):
        unioned = lines[0].overlay(lines[1], how="union", keep_geom_type=True)
    else:
        unioned = pd.concat([lines[0], lines[1]], ignore_index=True)

    if len(lines) > 2:
        for line_gdf in lines[2:]:
            if len(line_gdf):
                unioned = unioned.overlay(line_gdf, how="union", keep_geom_type=True)
            else:
                unioned = pd.concat([unioned, line_gdf], ignore_index=True)

    return make_all_singlepart(unioned, ignore_index=True)


def clean_clip(
    gdf: GeoDataFrame | GeoSeries,
    mask: GeoDataFrame | GeoSeries | Geometry,
    keep_geom_type: bool | None = None,
    geom_type: str | None = None,
    **kwargs,
) -> GeoDataFrame | GeoSeries:
    """Clips and clean geometries.

    Geopandas.clip does a "fast and dirty clipping, with no guarantee for valid
    outputs". Here, the clipped geometries are made valid, and empty and NaN
    geometries are removed.

    Args:
        gdf: GeoDataFrame or GeoSeries to be clipped
        mask: the geometry to clip gdf
        geom_type: Optionally specify what geometry type to keep.,
            if there are mixed geometry types. Must be either "polygon",
            "line" or "point".
        keep_geom_type: Defaults to None, meaning True if 'geom_type' is given
            and True if the geometries are single-typed and False if the geometries
            are mixed.
        **kwargs: Keyword arguments passed to geopandas.GeoDataFrame.clip

    Returns:
        The cleanly clipped GeoDataFrame.

    Raises:
        TypeError: If gdf is not of type GeoDataFrame or GeoSeries.
    """
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    gdf, geom_type, keep_geom_type = _determine_geom_type_args(
        gdf, geom_type, keep_geom_type
    )

    try:
        gdf = gdf.clip(mask, **kwargs).pipe(clean_geoms)
    except Exception:
        gdf = clean_geoms(gdf)
        try:
            mask = clean_geoms(mask)
        except TypeError:
            mask = make_valid(mask)

        return gdf.clip(mask, **kwargs).pipe(clean_geoms)

    if keep_geom_type:
        gdf = to_single_geom_type(gdf, geom_type)

    return gdf


def _determine_geom_type_args(
    gdf: GeoDataFrame, geom_type: str | None, keep_geom_type: bool | None
) -> tuple[GeoDataFrame, str, bool]:
    if geom_type:
        gdf = to_single_geom_type(gdf, geom_type)
        keep_geom_type = True
    elif keep_geom_type is None:
        geom_type = get_geom_type(gdf)
        if geom_type == "mixed":
            keep_geom_type = False
        else:
            keep_geom_type = True
    elif keep_geom_type:
        geom_type = get_geom_type(gdf)
        if geom_type == "mixed":
            raise ValueError("Cannot set keep_geom_type=True with mixed geometries")
    return gdf, geom_type, keep_geom_type


def _merge_geometries(geoms: GeoSeries, grid_size=None) -> Geometry:
    return make_valid(unary_union(geoms, grid_size=grid_size))


def _parallel_unary_union(
    gdf: GeoDataFrame, n_jobs: int = 1, by=None, grid_size=None, **kwargs
) -> list[Geometry]:
    try:
        geom_col = gdf._geometry_column_name
    except AttributeError:
        geom_col = "geometry"

    if by is not None and not isinstance(by, str):
        gdf = gdf.copy()
        try:
            gdf["_by"] = gdf[by].astype(str).agg("-".join, axis=1)
        except KeyError:
            gdf["_by"] = by
        by = "_by"

    if gdf.crs is None:
        gdf.crs = 25833
        _was_none = True
    else:
        _was_none = False

    if isinstance(gdf.index, pd.MultiIndex):
        gdf = gdf.reset_index(drop=True)

    dissolved = (
        dask_geopandas.from_geopandas(gdf, npartitions=n_jobs).dissolve(by).compute()
    )
    if _was_none:
        dissolved.crs = None

    return dissolved.geometry


def _parallel_unary_union_geoseries(
    ser: GeoSeries, n_jobs: int = 1, grid_size=None, **kwargs
) -> list[Geometry]:
    if ser.crs is None:
        ser.crs = 25833
        _was_none = True
    else:
        _was_none = False

    if isinstance(ser.index, pd.MultiIndex):
        ser = ser.reset_index(drop=True)

    dissolved = (
        dask_geopandas.from_geopandas(ser.to_frame("geometry"), npartitions=n_jobs)
        .dissolve(**kwargs)
        .compute()
    )
    if _was_none:
        dissolved.crs = None

    return dissolved.geometry


def _parallel_unary_union(
    gdf: GeoDataFrame, n_jobs: int = 1, by=None, grid_size=None, **kwargs
) -> list[Geometry]:
    try:
        geom_col = gdf._geometry_column_name
    except AttributeError:
        geom_col = "geometry"

    with joblib.Parallel(n_jobs=n_jobs, backend="threading") as parallel:
        delayed_operations = []
        for _, geoms in gdf.groupby(by, **kwargs)[geom_col]:
            delayed_operations.append(
                joblib.delayed(_merge_geometries)(geoms, grid_size=grid_size)
            )

        return parallel(delayed_operations)


def _parallel_unary_union_geoseries(
    ser: GeoSeries, n_jobs: int = 1, grid_size=None, **kwargs
) -> list[Geometry]:

    is_one_hit = ser.groupby(**kwargs).transform("size") == 1

    one_hit = ser.loc[is_one_hit]
    many_hits = ser.loc[~is_one_hit]

    with joblib.Parallel(n_jobs=n_jobs, backend="threading") as parallel:
        delayed_operations = []
        for _, geoms in many_hits.groupby(**kwargs):
            delayed_operations.append(
                joblib.delayed(_merge_geometries)(geoms, grid_size=grid_size)
            )

        dissolved = pd.Series(
            parallel(delayed_operations),
            index=is_one_hit[lambda x: x is False].index.unique(),
        )

    return pd.concat([dissolved, one_hit]).sort_index().values


def _parallel_unary_union_geoseries(
    ser: GeoSeries, n_jobs: int = 1, grid_size=None, **kwargs
) -> list[Geometry]:

    with joblib.Parallel(n_jobs=n_jobs, backend="threading") as parallel:
        delayed_operations = []
        for _, geoms in ser.groupby(**kwargs):
            delayed_operations.append(
                joblib.delayed(_merge_geometries)(geoms, grid_size=grid_size)
            )

        return parallel(delayed_operations)



import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from igraph import Graph
from shapely import force_2d
from shapely import reverse
from shapely import unary_union
from shapely.geometry import MultiPoint
from shapely.geometry import Point
from shapely.ops import nearest_points
from shapely.wkt import loads

from ..geopandas_tools.sfilter import sfilter
from .cutting_lines import cut_lines_once
from .nodes import make_edge_wkt_cols


def _service_area(
    graph: Graph,
    origins: GeoDataFrame,
    weight: str,
    breaks: np.ndarray,
    lines: GeoDataFrame,
    nodes: GeoDataFrame,
    directed: bool,
    precice: bool,
) -> GeoDataFrame:
    # make sure the nodes are alligned with the vertices in the graph
    # the weight/distances can then be assigned directly onto the nodes
    node_df = pd.DataFrame(index=np.array(graph.vs["name"]))
    nodes = nodes.set_index("node_id").drop("geometry", axis=1)
    nodes = node_df.join(nodes)

    # double edge df with source/target as index
    edge_df = lines.loc[:, ["source", "target", "src_tgt_wt", "geometry"]]
    edge_df = pd.concat(
        [
            edge_df.set_index("source", drop=False),
            edge_df.set_index("target", drop=False),
        ]
    )

    # distance from all origins to all vertices/nodes in the graph
    all_distances: list[list[str]] = graph.distances(
        weights="weight", source=origins["temp_idx"], mode="out"
    )

    if not directed:
        nodes_union: MultiPoint = unary_union(loads(nodes["wkt"].dropna().values))

    # loop through every origin and every break
    service_areas: list[GeoDataFrame] = []
    for i, idx in enumerate(origins["temp_idx"]):
        # assign distances to the nodes and join the column to the edges
        nodes[weight] = all_distances[i]
        distance_df = edge_df.join(nodes)

        for break_ in breaks:
            nodes_within_break = nodes.loc[nodes[weight] <= break_]

            whole_edge_is_within = (edge_df.source.isin(nodes_within_break.index)) & (
                edge_df.target.isin(nodes_within_break.index)
            )
            edges_within = edge_df.loc[whole_edge_is_within]

            if not precice:
                if not len(edges_within):
                    edges_within = GeoDataFrame(
                        {"geometry": [None]}, geometry="geometry", crs=lines.crs
                    )
                edges_within["origin"] = idx
                edges_within[weight] = break_
                service_areas.append(edges_within)
                continue

            # only part of the line is within the break
            partly_within = _part_of_edge_within(
                distance_df, nodes_within_break, directed
            )

            if not len(edges_within) and not len(partly_within):
                nan_gdf = GeoDataFrame(
                    {"origin": [idx], weight: [break_], "geometry": [None]},
                    geometry="geometry",
                    crs=lines.crs,
                )
                service_areas.append(nan_gdf)
                continue

            if not len(partly_within):
                edges_within["origin"] = idx
                edges_within[weight] = break_
                service_areas.append(edges_within)
                continue

            partly_within["remaining_distance"] = break_ - partly_within[weight]

            split_lines = _split_lines(partly_within, directed)

            split_lines = make_edge_wkt_cols(split_lines)

            # select the cutted lines shorter than the cut distance that intersects
            # with the nodes
            within = split_lines.loc[
                lambda df: (df.length <= df["remaining_distance"] * 1.01)
                & (df["source_wkt"].isin(nodes.wkt) | df["target_wkt"].isin(nodes.wkt))
            ]

            # keep only lines intersecting the edges completely within or the snapped origin
            if not directed:
                if len(edges_within):
                    within = sfilter(within, edges_within.buffer(0.01))
                else:
                    snapped_origin: Point = nearest_points(
                        nodes_union,
                        origins.loc[origins["temp_idx"] == idx, "geometry"].unary_union,
                    )[0]

                    within = sfilter(within, snapped_origin.buffer(0.01))

            edges_within = pd.concat([edges_within, within], ignore_index=True)
            edges_within["origin"] = idx
            edges_within[weight] = break_

            service_areas.append(edges_within)

    return pd.concat(
        service_areas,
        ignore_index=True,
    )


def _part_of_edge_within(distance_df, nodes_within_break, directed):
    if directed:
        return distance_df.loc[(distance_df.source.isin(nodes_within_break.index))]
    else:
        return distance_df.loc[
            (distance_df.source.isin(nodes_within_break.index))
            | (distance_df.target.isin(nodes_within_break.index))
        ]


def _split_lines(partly_within, directed):
    partly_within.geometry = force_2d(partly_within.geometry)

    if not directed:
        rev = partly_within.copy()
        rev.geometry = reverse(partly_within.geometry)
        partly_within = pd.concat([partly_within, rev])

    lines_cut = cut_lines_once(partly_within, "remaining_distance")

    return lines_cut



from collections.abc import Callable
from typing import Any

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from geopandas.array import GeometryArray
from numpy.typing import NDArray
from pyproj import CRS
from shapely import get_coordinates
from shapely import get_exterior_ring
from shapely import get_interior_ring
from shapely import get_num_interior_rings
from shapely import linearrings
from shapely import make_valid
from shapely import polygons
from shapely import unary_union
from shapely.geometry import LinearRing
from shapely.geometry import Polygon

from .conversion import to_gdf
from .conversion import to_geoseries


class PolygonsAsRings:
    """Convert polygons to linearrings, apply linestring functions, then convert back to polygons."""

    def __init__(
        self,
        polys: GeoDataFrame | GeoSeries | GeometryArray,
        crs: CRS | Any | None = None,
        allow_multipart: bool = False,
        gridsize: int | None = None,
    ) -> None:
        """Initialize the PolygonsAsRings object with polygons and optional CRS information.

        Args:
            polys: GeoDataFrame, GeoSeries, or GeometryArray containing polygon geometries.
            crs: Coordinate Reference System to be used, defaults to None.
            allow_multipart: Allow multipart polygons if True, defaults to False.
            gridsize: Size of the grid for any grid operations, defaults to None.
        """
        if not isinstance(polys, (pd.DataFrame, pd.Series, GeometryArray)):
            raise TypeError(type(polys))

        self.polyclass = polys.__class__

        if not isinstance(polys, pd.DataFrame):
            polys = to_gdf(polys, crs)

        if not allow_multipart and not (polys.geom_type == "Polygon").all():
            raise ValueError(
                "All geometries must be single-type Polygons. Set allow_multipart=True to allow MultiPolygons",
                polys.geom_type.value_counts(),
            )
        if not polys.geom_type.isin(["Polygon", "MultiPolygon"]).all():
            raise ValueError(
                f"All geometries must be Polygons. Got {polys.geom_type.value_counts()}"
            )

        self._index_mapper = dict(enumerate(polys.index))
        self.gdf = polys.reset_index(drop=True)

        if crs is not None:
            self.crs = crs
        elif hasattr(polys, "crs"):
            self.crs = polys.crs
        else:
            self.crs = None

        if not len(self.gdf):
            self.rings = pd.Series()
            return

        exterior = pd.Series(
            get_exterior_ring(self.gdf.geometry.values),
            index=self._exterior_index,
        )

        self.max_rings: int = np.max(get_num_interior_rings(self.gdf.geometry.values))

        if not self.max_rings:
            self.rings = exterior
            return

        # series same length as number of potential inner rings
        interiors = pd.Series(
            (
                [
                    [get_interior_ring(geom, i) for i in range(self.max_rings)]
                    for geom in self.gdf.geometry
                ]
            ),
        ).explode()

        interiors.index = self._interiors_index

        interiors = interiors.dropna()

        self.rings = pd.concat([exterior, interiors])

    def get_rings(self, agg: bool = False) -> GeoDataFrame | GeoSeries | np.ndarray:
        """Retrieve rings from the polygons, optionally aggregating them.

        Args:
            agg: If True, aggregate the rings into single geometries.

        Returns:
            The rings either aggregated or separated, in the type of
                the input polygons.
        """
        gdf = self.gdf.copy()
        rings = self.rings.copy()
        if not len(rings):
            return GeoSeries()
        if agg:
            gdf.geometry = rings.groupby(level=1).agg(unary_union)
        else:
            rings.index = rings.index.get_level_values(1)
            rings.name = "geometry"
            gdf = gdf.drop(columns="geometry").join(rings)

        if issubclass(self.polyclass, pd.DataFrame):
            return GeoDataFrame(gdf, crs=self.crs)
        if issubclass(self.polyclass, pd.Series):
            return GeoSeries(gdf.geometry)
        return self.polyclass(gdf.geometry.values)

    def apply_numpy_func_to_interiors(
        self,
        func: Callable,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> "PolygonsAsRings":
        """Apply a numpy function specifically to the interior rings of the polygons.

        Args:
            func: Numpy function to apply.
            args: Tuple of positional arguments for the function.
            kwargs: Dictionary of keyword arguments for the function.

        Returns:
            PolygonsAsRings: The instance itself after applying the function.
        """
        kwargs = kwargs or {}
        args = args or ()
        arr: NDArray[LinearRing] = self.rings.loc[self.is_interior].values
        index: pd.Index = self.rings.loc[self.is_interior].index
        results = pd.Series(
            np.array(func(arr, *args, **kwargs)),
            index=index,
        )
        self.rings.loc[self.is_interior] = results
        return self

    def apply_numpy_func(
        self,
        func: Callable,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> "PolygonsAsRings":
        """Apply a numpy function to all rings of the polygons.

        Args:
            func: Numpy function to apply.
            args: Tuple of positional arguments for the function.
            kwargs: Dictionary of keyword arguments for the function.

        Returns:
            PolygonsAsRings: The instance itself after applying the function.
        """
        kwargs = kwargs or {}
        args = args or ()

        results = np.array(func(self.rings.values, *args, **kwargs))

        if len(results) != len(self.rings):
            raise ValueError(
                f"Different length of results. Got {len(results)} and {len(self.rings)} original rings"
            )

        self.rings.loc[:] = results  # type: ignore [call-overload]

        return self

    def apply_geoseries_func(
        self,
        func: Callable,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> "PolygonsAsRings":
        """Apply a function that operates on a GeoSeries to the rings.

        Args:
            func: Function to apply that expects a GeoSeries.
            args: Tuple of positional arguments for the function.
            kwargs: Dictionary of keyword arguments for the function.

        Returns:
            PolygonsAsRings: The instance itself after applying the function.
        """
        kwargs = kwargs or {}
        args = args or ()

        self.rings.loc[:] = np.array(  # type: ignore [call-overload]
            func(
                GeoSeries(
                    self.rings.values,
                    crs=self.crs,
                    index=self.rings.index.get_level_values(1).map(self._index_mapper),
                ),
                *args,
                **kwargs,
            )
        )

        return self

    def apply_gdf_func(
        self,
        func: Callable,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> "PolygonsAsRings":
        """Apply a function that operates on a GeoDataFrame to the rings.

        Args:
            func: Function to apply that expects a GeoDataFrame.
            args: Tuple of positional arguments for the function.
            kwargs: Dictionary of keyword arguments for the function.

        Returns:
            PolygonsAsRings: The instance itself after applying the function.
        """
        kwargs = kwargs or {}
        args = args or ()

        gdf = GeoDataFrame(
            {"geometry": self.rings.values},
            crs=self.crs,
            index=self.rings.index.get_level_values(1).map(self._index_mapper),
        ).join(self.gdf.drop(columns="geometry"))

        assert len(gdf) == len(self.rings)

        gdf.index = self.rings.index

        self.rings.loc[:] = func(  # type: ignore [call-overload]
            gdf,
            *args,
            **kwargs,
        ).geometry.values

        return self

    @property
    def is_interior(self) -> bool:
        """Returns a boolean Series of whether the row is an interior ring."""
        return self.rings.index.get_level_values(0) == 1

    @property
    def is_exterior(self) -> bool:
        """Returns a boolean Series of whether the row is an exterior ring."""
        return self.rings.index.get_level_values(0) == 0

    @property
    def _interiors_index(self) -> pd.MultiIndex:
        """A three-leveled MultiIndex.

        Used to separate interior and exterior and sort the interior in
        the 'to_numpy' method.

        level 0: all 1s, indicating "is interior".
        level 1: gdf index repeated *self.max_rings* times.
        level 2: interior number index. 0 * len(gdf), 1 * len(gdf), 2 * len(gdf)...
        """
        if not self.max_rings:
            return pd.MultiIndex()
        len_gdf = len(self.gdf)
        n_potential_interiors = len_gdf * self.max_rings
        gdf_index = sorted(list(self.gdf.index) * self.max_rings)
        interior_number_index = np.tile(np.arange(self.max_rings), len_gdf)
        one_for_interior = np.repeat(1, n_potential_interiors)

        return pd.MultiIndex.from_arrays(
            [one_for_interior, gdf_index, interior_number_index]
        )

    @property
    def _exterior_index(self) -> pd.MultiIndex:
        """A three-leveled MultiIndex.

        Used to separate interior and exterior in the 'to_numpy' method.
        Only leve 1 is used for the exterior.

        level 0: all 0s, indicating "not interior".
        level 1: gdf index.
        level 2: All 0s.
        """
        zero_for_not_interior = np.repeat(0, len(self.gdf))
        return pd.MultiIndex.from_arrays(
            [zero_for_not_interior, self.gdf.index, zero_for_not_interior]
        )

    def to_gdf(self) -> GeoDataFrame:
        """Return the GeoDataFrame with polygons."""
        self.gdf.geometry = self.to_numpy()
        return self.gdf

    def to_geoseries(self) -> GeoSeries:
        """Return the GeoSeries with polygons."""
        self.gdf.geometry = self.to_numpy()
        return self.gdf.geometry

    def to_numpy(self) -> NDArray[Polygon]:
        """Return a numpy array of polygons."""
        if not len(self.rings):
            return np.array([])

        exterior = self.rings.loc[self.is_exterior].sort_index()
        assert exterior.shape == (len(self.gdf),)

        nonempty_interiors = self.rings.loc[self.is_interior]

        if not len(nonempty_interiors):
            try:
                return make_valid(polygons(exterior.values))
            except Exception:
                return _geoms_to_linearrings_fallback(exterior).values

        empty_interiors = pd.Series(
            [None for _ in range(len(self.gdf) * self.max_rings)],
            index=self._interiors_index,
        ).loc[lambda x: ~x.index.isin(nonempty_interiors.index)]

        interiors = (
            pd.concat([nonempty_interiors, empty_interiors])
            .sort_index()
            # make each ring level a column with same length and order as gdf
            .unstack(level=2)
            .sort_index()
        )
        assert interiors.shape == (len(self.gdf), self.max_rings), interiors.shape

        try:
            return make_valid(polygons(exterior.values, interiors.values))
        except Exception:
            return _geoms_to_linearrings_fallback(exterior, interiors).values


def get_linearring_series(geoms: GeoDataFrame | GeoSeries) -> pd.Series:
    """Convert geometries into a series of LinearRings.

    Args:
        geoms: GeoDataFrame or GeoSeries from which to extract LinearRings.

    Returns:
        pd.Series: A series containing LinearRings.
    """
    geoms = to_geoseries(geoms).explode(index_parts=False)
    coords, indices = get_coordinates(geoms, return_index=True)
    return pd.Series(linearrings(coords, indices=indices), index=geoms.index)


def _geoms_to_linearrings_fallback(
    exterior: pd.Series, interiors: pd.Series | None = None
) -> pd.Series:
    exterior.index = exterior.index.get_level_values(1)
    assert exterior.index.is_monotonic_increasing

    exterior = get_linearring_series(exterior)

    if interiors is None:
        return (
            pd.Series(
                make_valid(polygons(exterior.values)),
                index=exterior.index,
            )
            .groupby(level=0)
            .agg(lambda x: make_valid(unary_union(x)))
        )

    interiors.index = interiors.index.get_level_values(1)
    new_interiors = []
    for col in interiors:
        new_interiors.append(get_linearring_series(interiors[col]))

    all_none = [[None] * len(new_interiors)] * len(exterior)
    cols = list(interiors.columns)
    out_interiors = pd.DataFrame(
        all_none,
        columns=cols,
        index=exterior.index,
    )
    out_interiors[cols] = pd.concat(new_interiors, axis=1)
    for col in out_interiors:
        out_interiors.loc[out_interiors[col].isna(), col] = None

    return (
        pd.Series(
            make_valid(polygons(exterior.values, out_interiors.values)),
            index=exterior.index,
        )
        .groupby(level=0)
        .agg(lambda x: make_valid(unary_union(x)))
    )



from xyzservices import Bunch
from xyzservices import TileProvider
from xyzservices import providers

kartverket = Bunch(
    norgeskart=TileProvider(
        name="Norgeskart",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=norgeskart_bakgrunn&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    bakgrunnskart_forenklet=TileProvider(
        name="Norgeskart forenklet",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=bakgrunnskart_forenklet&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    norges_grunnkart=TileProvider(
        name="Norges grunnkart",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=norges_grunnkart&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    norges_grunnkart_gråtone=TileProvider(
        name="Norges grunnkart gråtone",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=norges_grunnkart_graatone&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    n50=TileProvider(
        name="N5 til N50 kartdata",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=kartdata3&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    topogråtone=TileProvider(
        name="Topografisk norgeskart gråtone",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=topo4graatone&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    toporaster=TileProvider(
        name="Topografisk raster",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=toporaster4&zoom={z}&x={x}&y={y}",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    norge_i_bilder=TileProvider(
        name="Norge i bilder",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_nib_web_mercator_wmts_v2?SERVICE=WMTS&REQUEST=GetTile&VERSION=1.0.0&LAYER=Nibcache_web_mercator_v2&STYLE=default&FORMAT=image/jpgpng&tileMatrixSet=default028mm&tileMatrix={z}&tileRow={y}&tileCol={x}",
        max_zoom=19,
        attribution="© Geovekst",
    ),
)

xyz = Bunch({"Kartverket": kartverket} | providers)



from collections.abc import Callable

import numpy as np
import pandas as pd

from .raster import Raster


def ndvi(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    # normalize red and nir arrays to 0-1 scale if needed
    if red.max() > 1 and nir.max() > 1:
        red = red / 255
        nir = nir / 255
    elif red.max() > 1 or nir.max() > 1:
        raise ValueError()

    ndvi_values = np.where((red + nir) == 0, 0, (nir - red) / (nir + red))

    return ndvi_values


def gndvi(green: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return np.where((green + nir) == 0, 0, (nir - green) / (nir + green))


def water(green: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return np.where((green + nir) == 0, 0, (green - nir) / (green + nir))


def builtup(swir: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return np.where((swir + nir) == 0, 0, (swir - nir) / (swir + nir))


def moisture(swir: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return np.where((swir + nir) == 0, 0, (nir - swir) / (nir + swir))


def get_raster_pairs(
    cube,
    band_name1: str,
    band_name2: str,
) -> list[tuple[Raster, Raster]]:
    unique = pd.DataFrame({"tile": cube.tile, "date": cube.date}).drop_duplicates(
        ["tile", "date"]
    )

    raster_pairs = []
    for tile, date in zip(unique["tile"], unique["date"], strict=False):
        query = (cube.tile == tile) & (cube.date == date)
        band1 = cube.copy()[query & (cube.band == band_name1)]
        band2 = cube.copy()[query & (cube.band == band_name2)]
        if not len(band1) and not len(band2):
            continue
        if len(band1) > 1:
            raise ValueError("Cannot have more than one B4 band per tile.")
        if len(band2) > 1:
            raise ValueError("Cannot have more than one B8 band per tile.")
        if len(band1) != 1 and len(band2) != 1:
            raise ValueError("Must have one B4 and one B8 band per tile.")
        band1 = band1[0]
        band2 = band2[0]
        assert isinstance(band1, Raster), band1
        assert isinstance(band2, Raster), band2

        if band1.shape != band2.shape:
            raise ValueError("Rasters must have same shape")

        pair = band1, band2
        raster_pairs.append(pair)

    return raster_pairs


def index_calc_pair(
    raster_pair: tuple[Raster, Raster], index_formula: Callable
) -> Raster:
    """Calculate an index for one raster pair and return a single Raster."""
    r1, r2 = raster_pair
    assert isinstance(r1, Raster), r1
    assert isinstance(r2, Raster), r2

    if r1.values is None:
        r1 = r1.load()
    if r2.values is None:
        r2 = r2.load()

    r1_arr: np.ndarray = r1.values.astype(np.float16)
    r2_arr: np.ndarray = r2.values.astype(np.float16)

    out_array = index_formula(r1_arr, r2_arr)

    return Raster.from_array(
        out_array,
        crs=r1.crs,
        bounds=r1.bounds,
        name=index_formula.__name__,
        date=r1.date,
    )



import networkx as nx
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from networkx.utils import pairwise
from shapely.geometry import Point

from ..geopandas_tools.conversion import to_geoseries
from ..geopandas_tools.neighbors import get_all_distances


def traveling_salesman_problem(
    points: GeoDataFrame | GeoSeries,
    return_to_start: bool = True,
    distances: pd.DataFrame | None = None,
) -> list[Point]:
    """Get the shortest path visiting all points.

    Args:
        points: An iterable of point geometries.
        return_to_start: If True (default), the path
            will make a full circle to the startpoint.
            If False, a dummy node will be added to make the
            salesman focus only on getting to the last node. Not
            guaranteed to work, meaning the wrong edge might be removed.
        distances: Optional DataFrame of distances between all points.
            If not provided, the calculation is done within this function.
            The DataFrame should be identical to the DataFrame created
            from sgis.get_all_distances from and to the points.

    Returns:
        List of Points making up the traveling salesman's path.

    Examples:
    ---------
    >>> import sgis as sg
    >>> from shapely.geometry import LineString
    >>> points = sg.to_gdf(
    ...     [
    ...         (0, 0),
    ...         (10, -10),
    ...         (10, 10),
    ...         (0, 10),
    ...         (0, -10),
    ...         (10, 0),
    ...         (20, 0),
    ...         (0, 20),
    ...     ]
    ... )
    >>> roundtrip = sg.traveling_salesman_problem(points)
    >>> roundtrip
    [<POINT (0 0)>, <POINT (10 -10)>, <POINT (0 -10)>, <POINT (10 0)>, <POINT (20 0)>, <POINT (10 10)>, <POINT (0 10)>, <POINT (0 20)>, <POINT (0 0)>]

    >>> LineString(roundtrip)
    LINESTRING (0 0, 10 -10, 0 -10, 10 0, 20 0, 10 10, 0 10, 0 20, 0 0)

    >>> oneway_trip = sg.traveling_salesman_problem(points, return_to_start=False)
    >>> oneway_trip
    [<POINT (0 0)>, <POINT (10 0)>, <POINT (20 0)>, <POINT (10 -10)>, <POINT (0 -10)>, <POINT (10 10)>, <POINT (0 10)>, <POINT (0 0)>]

    >>> LineString(oneway_trip)
    LINESTRING (0 20, 0 10, 10 10, 0 0, 10 0, 20 0, 10 -10, 0 -10)

    """
    points = to_geoseries(points).drop_duplicates()

    if len(points) <= 2:
        return list(points.dropna())

    if distances is None:
        points.index = range(len(points))
        distances: pd.DataFrame = get_all_distances(points, points)
    else:
        if not points.index.is_unique:
            raise ValueError("Index must be unique when passing 'distances'")

        distances = distances.loc[
            lambda x: (x.index.isin(points.index))
            & (x["neighbor_index"].isin(points.index))
        ]

        # need tange integer index
        to_int_idx = {idx: i for i, idx in enumerate(points.index)}
        points.index = points.index.map(to_int_idx)
        points = points.sort_index()
        distances.index = distances.index.map(to_int_idx)
        distances["neighbor_index"] = distances["neighbor_index"].map(to_int_idx)

    idx_to_point: dict[int, Point] = dict(enumerate(points))

    if not return_to_start:
        distances["mean_distance"] = distances.groupby(level=0)["distance"].transform(
            "mean"
        )

        distances = distances.sort_values(
            ["mean_distance", "distance"], ascending=[True, False]
        )

        max_dist_idx = distances["mean_distance"].idxmax()

        dummy_node_idx = points.index.max() + 1
        n_points = dummy_node_idx + 1
        max_dist_and_some = distances["distance"].sum() * 1.01

        # add edges in both directions to the dummy node
        dummy_node = pd.DataFrame(
            {
                "neighbor_index": [i for i in range(n_points)]
                + [dummy_node_idx] * dummy_node_idx,
                "distance": [max_dist_and_some for _ in range(n_points * 2 - 1)],
            },
            index=[dummy_node_idx] * (n_points) + [i for i in range(dummy_node_idx)],
        )

        dummy_node.loc[
            lambda x: (x["neighbor_index"] == max_dist_idx)
            | (x.index == max_dist_idx)
            | (x["neighbor_index"] == x.index),
            "distance",
        ] = 0

        distances = pd.concat([distances, dummy_node])
    else:
        n_points = points.index.max() + 1

    # now to mimick the return values of nx.all_pairs_dijkstra, nested dictionaries of distances and nodes/edges
    dist, path = {}, {}
    for i in distances.index.unique():
        dist[i] = dict(distances.loc[i, ["neighbor_index", "distance"]].values)
        path[i] = {
            neighbor: [i, neighbor] for neighbor in distances.loc[i, "neighbor_index"]
        }

    # the rest of the function is copied from networkx' traveling_salesman_problem

    nx_graph = nx.Graph()
    for u in range(n_points):
        for v in range(n_points):
            if u == v:
                continue
            nx_graph.add_edge(u, v, weight=dist[u][v])
    best = nx.approximation.christofides(nx_graph, "weight")

    best_path = []
    for u, v in pairwise(best):
        best_path.extend(path[u][v][:-1])
    best_path.append(v)

    if return_to_start:
        return [idx_to_point[i] for i in best_path]

    # drop duplicates, but keep order
    best_path = list(dict.fromkeys(best_path))

    idx_start = best_path.index(dummy_node_idx)  # - 1

    best_path = best_path[idx_start:] + best_path[:idx_start]

    as_points = [idx_to_point[i] for i in best_path if i != dummy_node_idx]

    return as_points  # + [as_points[0]]



"""Functions for point geometries."""

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely import distance
from shapely import unary_union
from shapely.ops import nearest_points

from ..geopandas_tools.geometry_types import get_geom_type
from ..geopandas_tools.geometry_types import to_single_geom_type
from ..geopandas_tools.polygon_operations import PolygonsAsRings


def snap_within_distance(
    points: GeoDataFrame | GeoSeries,
    to: GeoDataFrame | GeoSeries,
    max_distance: int | float,
    *,
    distance_col: str | None = None,
) -> GeoDataFrame | GeoSeries:
    """Snaps points to nearest geometry if within given distance.

    It takes a GeoDataFrame of points and snaps them to the nearest geometry in a
    second GeoDataFrame if the snap distance is less than 'max_distance'.
    Adds a distance column if specified.

    Args:
        points: The GeoDataFrame of points to snap.
        to: The GeoDataFrame to snap to.
        max_distance: The maximum distance to snap to.
        distance_col: Name of column with the snap distance. Defaults to
            'snap_distance'. Set to None to not get any distance column. This will make
            the function a bit faster.

    Returns:
        A GeoDataFrame or GeoSeries with the points snapped to the nearest point in the
        'to' GeoDataFrame or GeoSeries.

    Notes:
        If there are  geometries equally close to the points, one geometry will be
        chosen as the snap geometry. This will usually only happen with constructed
        data like grids or in the examples below.

        The snap point might be in between vertices of lines and polygons. Convert the
        'to' geometries to multipoint before snapping if the snap points should be
        vertices.

    Examples:
    ---------
    Create som points.

    >>> from sgis import snap_within_distance, to_gdf
    >>> points = to_gdf([(0, 0), (1, 1)])
    >>> points
                    geometry
    0  POINT (0.00000 0.00000)
    1  POINT (1.00000 1.00000)
    >>> to = to_gdf([(2, 2), (3, 3)])
    >>> to
                    geometry
    0  POINT (2.00000 2.00000)
    1  POINT (3.00000 3.00000)

    Snap 'points' to closest geometry in 'to' if distance is less 1, 2 and 3.

    >>> snap_within_distance(points, to, 1)
                    geometry snap_distance
    0  POINT (0.00000 0.00000)          <NA>
    1  POINT (1.00000 1.00000)          <NA>
    >>> snap_within_distance(points, to, 2)
                        geometry snap_distance
    0  POINT (0.00000 0.00000)          <NA>
    1  POINT (2.00000 2.00000)      1.414214
    >>> snap_within_distance(points, to, 3)
                    geometry snap_distance
    0  POINT (2.00000 2.00000)      2.828427
    1  POINT (2.00000 2.00000)      1.414214
    """
    to = _polygons_to_rings(to)

    if not distance_col and not isinstance(points, GeoDataFrame):
        return _shapely_snap(
            points=points,
            to=to,
            max_distance=max_distance,
        )
    elif not isinstance(points, GeoDataFrame):
        points = points.to_frame()

    copied = points.copy()

    copied.geometry = _shapely_snap(
        points=copied.geometry.values,
        to=to,
        max_distance=max_distance,
    )

    if distance_col:
        copied[distance_col] = copied.distance(points)
        copied[distance_col] = np.where(
            copied[distance_col] == 0, pd.NA, copied[distance_col]
        )

    return copied


def snap_all(
    points: GeoDataFrame | GeoSeries,
    to: GeoDataFrame | GeoSeries,
    *,
    distance_col: str | None = None,
) -> GeoDataFrame | GeoSeries:
    """Snaps points to the nearest geometry.

    It takes a GeoDataFrame of points and snaps them to the nearest geometry in a
    second GeoDataFrame. Adds a distance column if specified.

    Args:
        points: The GeoDataFrame of points to snap.
        to: The GeoDataFrame to snap to.
        distance_col: Name of column with the snap distance. Defaults to None.

    Returns:
        A GeoDataFrame or GeoSeries with the points snapped to the nearest point in the
        'to' GeoDataFrame or GeoSeries.

    Notes:
        If there are  geometries equally close to the points, one geometry will be
        chosen as the snap geometry. This will usually only happen with constructed
        data like grids or in the examples below.

        The snap point might be in between vertices of lines and polygons. Convert the
        'to' geometries to multipoint before snapping if the snap points should be
        vertices.

    Examples:
    ---------
    Create som points.

    >>> from sgis import snap_all, to_gdf
    >>> points = to_gdf([(0, 0), (1, 1)])
    >>> points
                    geometry
    0  POINT (0.00000 0.00000)
    1  POINT (1.00000 1.00000)
    >>> to = to_gdf([(2, 2), (3, 3)])
    >>> to["snap_idx"] = to.index
    >>> to
                    geometry  snap_idx
    0  POINT (2.00000 2.00000)            0
    1  POINT (3.00000 3.00000)            1

    Snap all points to closest geometry in 'to'.

    >>> snap_all(points, to)
                    geometry  snap_distance
    0  POINT (2.00000 2.00000)       2.828427
    1  POINT (2.00000 2.00000)       1.414214
    """
    to = _polygons_to_rings(to)

    if not isinstance(points, GeoDataFrame):
        return _shapely_snap(
            points=points,
            to=to,
            max_distance=None,
        )

    copied = points.copy()

    copied.geometry = _shapely_snap(
        points=copied.geometry.values,
        to=to,
        max_distance=None,
    )

    if distance_col:
        copied[distance_col] = copied.distance(points)
        copied[distance_col] = np.where(
            copied[distance_col] == 0, pd.NA, copied[distance_col]
        )
    return copied


def _polygons_to_rings(gdf: GeoDataFrame) -> GeoDataFrame:
    if get_geom_type(gdf) == "polygon":
        return PolygonsAsRings(gdf).get_rings()
    if get_geom_type(gdf) != "mixed":
        return gdf
    gdf_points = to_single_geom_type(gdf, "point")
    gdf_lines = to_single_geom_type(gdf, "line")
    gdf_polys = PolygonsAsRings(to_single_geom_type(gdf, "polygon")).get_rings()
    return pd.concat([gdf_points, gdf_lines, gdf_polys])


def _shapely_snap(
    points: np.ndarray | GeoSeries,
    to: GeoSeries | GeoDataFrame,
    max_distance: int | float | None = None,
) -> GeoSeries:
    try:
        unioned = to.unary_union
    except AttributeError:
        unioned = unary_union(to)

    nearest = nearest_points(points, unioned)[1]

    if not max_distance:
        return nearest

    distances = distance(points, nearest)

    snapped = np.where(
        distances <= max_distance,
        nearest,
        points,
    )

    if isinstance(points, GeoSeries):
        return GeoSeries(points, crs=points.crs, index=points.index, name=points.name)

    return points.__class__(snapped)



import functools
import itertools
import warnings

import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from numpy.typing import NDArray
from shapely import STRtree
from shapely import distance
from shapely import extract_unique_points
from shapely import get_rings
from shapely import line_merge
from shapely import make_valid
from shapely import segmentize
from shapely import unary_union
from shapely import voronoi_polygons
from shapely.errors import GEOSException
from shapely.geometry import LineString
from shapely.ops import nearest_points

from ..maps.maps import explore
from ..networkanalysis.traveling_salesman import traveling_salesman_problem
from .conversion import to_gdf
from .conversion import to_geoseries
from .general import clean_geoms
from .general import make_lines_between_points
from .general import sort_long_first
from .geometry_types import make_all_singlepart
from .sfilter import sfilter_inverse
from .sfilter import sfilter_split

warnings.simplefilter(action="ignore", category=FutureWarning)


def get_traveling_salesman_lines(
    df: GeoDataFrame, return_to_start: bool = False
) -> list[LineString]:
    path = traveling_salesman_problem(df, return_to_start=return_to_start)

    try:
        return [LineString([p1, p2]) for p1, p2 in itertools.pairwise(path)]
    except IndexError as e:
        if len(path) == 1:
            return path
        raise e


def _remove_longest_if_not_intersecting(
    centerlines: GeoDataFrame, geoms: GeoDataFrame
) -> GeoDataFrame:
    centerlines = sort_long_first(make_all_singlepart(centerlines))

    has_only_one_line = centerlines.groupby(level=0).size() == 1
    only_one_line = centerlines[has_only_one_line]
    centerlines = centerlines[~has_only_one_line]

    longest = centerlines.loc[lambda x: ~x.index.duplicated()]
    not_longest = centerlines.loc[lambda x: x.index.duplicated()]

    longest_endpoints = longest.boundary.explode(index_parts=False).sort_index()

    nearest = longest_endpoints.groupby(level=0).apply(
        lambda x: nearest_points(
            x, not_longest[not_longest.index.isin(x.index)].unary_union
        )[1]
    )
    longest_endpoints.loc[:] = make_lines_between_points(
        longest_endpoints.values, nearest.values
    )

    return pd.concat([only_one_line, not_longest, longest_endpoints])


def get_rough_centerlines(
    gdf: GeoDataFrame,
    max_segment_length: int,
) -> GeoDataFrame:
    """Get a cheaply calculated centerline of a polygon.

    The line is not guaraneteed to be completely within the polygon. The line
    should start and end at the polygons' "endpoints".

    The function is meant for getting centerlines from slivers in coverage_clean and
    snap_polygons. It will give weird results and be extremely slow for
    complext polygons like (buffered) road networks.

    """
    precision = 0.01

    if not len(gdf):
        return gdf

    if not gdf.index.is_unique:
        raise ValueError("Index must be unique")

    geoms: GeoSeries = to_geoseries(gdf).explode(index_parts=False)

    segmentized: GeoSeries = segmentize(geoms, max_segment_length=max_segment_length)

    points: GeoSeries = _get_points_in_polygons(segmentized, precision)

    has_no_points = geoms.loc[(~geoms.index.isin(points.index))]

    more_points: GeoSeries = _get_points_in_polygons(
        has_no_points.buffer(precision), precision
    )

    # Geometries that have no lines inside, might be perfect circles.
    # These can get the centroid as centerline
    still_has_no_points = has_no_points.loc[
        (~has_no_points.index.isin(more_points.index))
    ]
    still_has_no_points.loc[:] = geoms.loc[still_has_no_points.index].centroid

    # very thin slivers
    has_points_now = has_no_points.loc[(has_no_points.index.isin(more_points.index))]

    if len(has_points_now):
        segmentized = pd.concat(
            [
                segmentized.loc[
                    ~segmentized.index.isin(
                        still_has_no_points.index.union(has_points_now.index)
                    )
                ],
                has_points_now,
            ]
        )
    else:
        segmentized = segmentized.loc[
            ~segmentized.index.isin(still_has_no_points.index)
        ]

    # make sure to include the endpoints
    endpoints = _get_approximate_polygon_endpoints(segmentized)

    geoms = geoms.loc[~geoms.index.isin(still_has_no_points.index)]

    # check if line between endpoints make up a decent centerline
    has_two = endpoints.groupby(level=0).size() == 2
    endpoint1 = endpoints.loc[has_two].groupby(level=0).first()
    endpoint2 = endpoints.loc[has_two].groupby(level=0).last()
    assert (endpoint1.index == endpoint2.index).all()

    end_to_end = GeoSeries(
        make_lines_between_points(endpoint1, endpoint2), index=endpoint1.index
    )

    # keep lines 90 percent intersecting the polygon
    length_now = end_to_end.length
    end_to_end = (
        end_to_end.intersection(geoms.buffer(precision))
        .dropna()
        .loc[lambda x: x.length > length_now * 0.9]
    )

    # straight end buffer to remove all in between ends
    to_be_erased = points.index.isin(end_to_end.index)

    dont_intersect = sfilter_inverse(
        points.iloc[to_be_erased], end_to_end.buffer(precision, cap_style=2)
    )

    points = (
        clean_geoms(
            pd.concat(
                [
                    points.iloc[~to_be_erased],
                    dont_intersect,
                ]
            )
        )
        .buffer(0.1)
        .groupby(level=0)
        .agg(unary_union)
        .explode(index_parts=False)
        .centroid
    )
    points = pd.concat(
        [
            points,
            endpoints,
        ]
    )

    explore(points=to_gdf(points, 25833), gdf=gdf)

    remove_longest = functools.partial(_remove_longest_if_not_intersecting, geoms=geoms)

    centerlines = GeoSeries(
        points.groupby(level=0).apply(get_traveling_salesman_lines).explode()
    ).pipe(remove_longest)

    # centerlines = sort_long_first(centerlines).loc[lambda x: x.index.duplicated()]

    # fix sharp turns by using the centroids of the centerline
    centerlines2 = GeoSeries(
        (
            pd.concat(
                [
                    centerlines.centroid,
                    endpoints,
                ]
            )
            .groupby(level=0)
            .apply(get_traveling_salesman_lines)
        ).explode()
    ).pipe(remove_longest)

    centerlines3 = GeoSeries(
        (
            pd.concat(
                [
                    centerlines2.centroid,
                    endpoints,
                ]
            )
            .groupby(level=0)
            .apply(get_traveling_salesman_lines)
        ).explode()
    ).pipe(remove_longest)

    centerlines = centerlines3.groupby(level=0).agg(
        lambda x: line_merge(unary_union(x))
        # lambda x: unary_union(x)
    )

    if isinstance(gdf, GeoSeries):
        return GeoSeries(
            pd.concat([centerlines, still_has_no_points]), crs=gdf.crs
        ).sort_index()

    centerlines = GeoDataFrame(
        {"geometry": pd.concat([centerlines, still_has_no_points])}, crs=gdf.crs
    )
    return centerlines


def _get_points_in_polygons(geometries: GeoSeries, precision: float) -> GeoSeries:
    # voronoi can cause problems if coordinates are nearly identical
    # buffering solves it
    try:
        voronoi_lines = voronoi_polygons(geometries, only_edges=True)
    except GEOSException:
        try:
            geometries = make_valid(geometries)
            voronoi_lines = voronoi_polygons(geometries, only_edges=True)
        except GEOSException:
            voronoi_lines = voronoi_polygons(
                geometries.buffer(precision).buffer(-precision), only_edges=True
            )

    crossing_lines = (
        geometries.buffer(precision, resolution=10)
        .intersection(voronoi_lines)
        .explode(index_parts=False)
    )
    within_polygons, not_within = sfilter_split(
        crossing_lines, geometries, predicate="within"
    )

    intersect_polys_at_two_places = (
        not_within.intersection(unary_union(get_rings(geometries))).geom_type
        == "MultiPoint"
    )
    not_within_but_relevant = not_within.loc[intersect_polys_at_two_places]

    return pd.concat([within_polygons, not_within_but_relevant]).centroid


def _get_approximate_polygon_endpoints(geoms: GeoSeries) -> GeoSeries:
    out_geoms = []

    are_thin = geoms.buffer(-1e-2).is_empty
    not_thin = geoms.loc[~are_thin]
    thin = geoms.loc[are_thin].buffer(1e-2)

    rectangles = pd.concat([not_thin, thin]).minimum_rotated_rectangle()

    # get_rings returns array with integer index that must be mapped to pandas index
    rings, indices = get_rings(rectangles, return_index=True)
    int_to_pd_index = dict(enumerate(rectangles.index))
    indices = [int_to_pd_index[i] for i in indices]

    rectangles.loc[:] = (
        pd.Series(rings, index=indices).groupby(level=0).agg(unary_union)
    )
    corner_points = (
        GeoSeries(
            extract_unique_points(rectangles)
            # adding a buffer, dissolve and explode to not get two corners in same end for very thin polygons
            .buffer(0.01)
            .groupby(level=0)
            .agg(unary_union)
        )
        .explode(index_parts=False)
        .centroid
    )

    nearest_ring_point = nearest_points(
        corner_points.values, geoms.loc[corner_points.index].values
    )[1]

    distance_to_corner = pd.Series(
        distance(nearest_ring_point, corner_points), index=corner_points.index
    )

    is_two_nearest: NDArray[bool] = (
        distance_to_corner.groupby(level=0)
        .apply(lambda x: x <= x.nsmallest(2).iloc[-1])
        .values
    )

    two_nearest = pd.Series(nearest_ring_point, index=corner_points.index).iloc[
        is_two_nearest
    ]

    # geometries with more than two endpoints now, are probably crosses/star-like
    more_than_two = two_nearest.loc[lambda x: x.groupby(level=0).size() > 2]
    if len(more_than_two):
        precision = rectangles.length / 100

        two_nearest = two_nearest.loc[lambda x: ~x.index.isin(more_than_two.index)]

        by_rectangle = (
            geoms.intersection(rectangles.buffer(precision))
            .explode(index_parts=False)
            .centroid
        )

        nearest_rectangle_points = nearest_points(by_rectangle, rectangles)[1]
        nearest_geom_points = nearest_points(nearest_rectangle_points, geoms)[1]

        out_geoms.append(nearest_geom_points)

    lines_around_geometries = _multipoints_to_line_segments(
        extract_unique_points(rectangles)
    )

    distance_to_rect = distance(
        two_nearest.values, rectangles.loc[two_nearest.index].values
    )

    # move the points to the rectangle
    to_be_moved = two_nearest.loc[distance_to_rect > 0.01]
    not_to_be_moved = two_nearest.loc[distance_to_rect <= 0.01]
    assert len(not_to_be_moved) + len(to_be_moved) == len(two_nearest)
    out_geoms.append(not_to_be_moved)

    if len(to_be_moved):
        tree = STRtree(lines_around_geometries.values)
        nearest_indices = tree.nearest(to_be_moved.values)

        relevant_lines_around_geometries = pd.Series(
            lines_around_geometries.iloc[nearest_indices].values,
            index=to_be_moved.index,
        )

        # then move the points to the closest vertice
        points_moved = pd.Series(
            nearest_points(
                relevant_lines_around_geometries.values,
                extract_unique_points(
                    geoms.loc[relevant_lines_around_geometries.index]
                ).values,
            )[1],
            index=to_be_moved.index,
        )
        out_geoms.append(points_moved)

    return pd.concat(out_geoms)


def _multipoints_to_line_segments(
    multipoints: GeoSeries | GeoDataFrame, to_next: bool = True, cycle: bool = True
) -> GeoSeries | GeoDataFrame:
    if not len(multipoints):
        return multipoints

    multipoints = to_geoseries(multipoints)

    if isinstance(multipoints.index, pd.MultiIndex):
        index = [
            multipoints.index.get_level_values(i)
            for i in range(multipoints.index.nlevels)
        ]
        multipoints.index = pd.MultiIndex.from_arrays(
            [list(range(len(multipoints))), *index],
            names=["range_idx", *multipoints.index.names],
        )
    else:
        multipoints.index = pd.MultiIndex.from_arrays(
            [np.arange(0, len(multipoints)), multipoints.index],
            names=["range_idx", multipoints.index.name],
        )

    try:
        crs = multipoints.crs
    except AttributeError:
        crs = None

    point_df = multipoints.explode(index_parts=False).to_frame("geometry")

    if to_next:
        shift = -1
        keep = "first"
    else:
        shift = 1
        keep = "last"

    point_df["next"] = point_df.groupby(level=0)["geometry"].shift(shift)

    if cycle:
        first_points: GeoSeries = point_df.loc[
            lambda x: ~x.index.get_level_values(0).duplicated(keep=keep), "geometry"
        ]
        is_last_point = point_df["next"].isna()

        point_df.loc[is_last_point, "next"] = first_points
        assert point_df["next"].notna().all()
    else:
        point_df = point_df[point_df["next"].notna()]

    point_df["geometry"] = [
        LineString([x1, x2])
        for x1, x2 in zip(point_df["geometry"], point_df["next"], strict=False)
    ]
    if isinstance(multipoints.index, pd.MultiIndex):
        point_df.index = point_df.index.droplevel(0)

    if isinstance(multipoints, GeoDataFrame):
        return GeoDataFrame(
            point_df.drop(columns=["next"]), geometry="geometry", crs=crs
        )
    return GeoSeries(point_df["geometry"], crs=crs)


def get_line_segments(
    lines: GeoDataFrame | GeoSeries, extract_unique: bool = False, cycle=False
) -> GeoDataFrame:
    try:
        assert lines.index.is_unique
    except AttributeError:
        pass

    lines = to_geoseries(lines)

    if extract_unique:
        points = extract_unique_points(lines.values)
    else:
        coords, indices = shapely.get_coordinates(lines, return_index=True)
        points = GeoSeries(shapely.points(coords), index=indices)

    return _multipoints_to_line_segments(points, cycle=cycle)



"""Function to check whether we are in Dapla without importing Dapla.

Function from: From https://github.com/statisticsnorway/ssb-altinn-python/blob/main/src/altinn/utils.py
"""

import os


def is_dapla() -> bool:
    """From https://github.com/statisticsnorway/ssb-altinn-python/blob/main/src/altinn/utils.py."""
    try:
        return os.environ["GCS_TOKEN_PROVIDER_KEY"] == "google"
    except KeyError:
        return False
